#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2024/12/16

import sys

from loguru import logger

if sys.platform == 'win32':
    try:  # 尝试加载VIP版
        from wxautox import WeChat
        from wxautox.elements import WxParam  # WxParam.DEFALUT_SAVEPATH 可以用来配置数据自动保存位置
    except ModuleNotFoundError:  # 否则用贫民版
        from wxauto import WeChat
        from wxauto.elements import WxParam

from pyxllib.prog.filelock import get_autogui_lock


class WeChatSingletonLock:
    """ 基于 get_autogui_lock 的微信全局唯一单例控制器，确保同一时间仅有一个微信自动化程序在操作 """

    def __init__(self, lock_timeout=-1, *, init=True):
        # 初始化全局锁
        self.lock = get_autogui_lock(timeout=lock_timeout)
        self.wx = WeChat() if init else None

    def __enter__(self):
        # 获取锁并激活微信窗口
        self.lock.acquire()
        if self.wx:
            self.wx._show()
            return self.wx

    def __exit__(self, exc_type, exc_value, traceback):
        # 释放锁
        self.lock.release()


def wechat_lock_send(user, text=None, files=None, url=None, *, timeout=-1, **kwargs):
    """ 使用全局唯一单例锁，确保同一时间仅有一个微信自动化程序在操作 """
    with WeChatSingletonLock(timeout) as we:
        # 241223周一12:27，今天可被这个默认2秒坑惨了，往错误群一直发骚扰消息
        # 22:07，但我复测，感觉不可能找不到啊，为什么会找到禅宗考勤管理群呢，太离谱了
        status = we.ChatWith(user, timeout=5)

        if status != user:
            raise ValueError(f'无法找到用户：{user}')

        if text:
            if kwargs.get('at') == '所有人':
                we.AtAll(text, user)
            else:
                we.SendMsg(text, user, **kwargs)
        if files:
            we.SendFiles(files, user, **kwargs)
        if url:
            we.SendUrlCard(url, user)


def wechat_handler(message):
    # 获取群名，如果没有指定，不使用此微信发送功能
    user = message.record["extra"].get("wechat_user")
    if user:
        wechat_lock_send(user, message)


if sys.platform == 'win32':
    # 创建专用的微信日志记录器，不绑定默认群名
    wechat_logger = logger.bind(wechat_user='文件传输助手')

    # 添加专用的微信处理器
    wechat_logger.add(wechat_handler,
                      format="{time:YYYY-MM-DD HH:mm:ss} | {level} | {name}:{function}:{line} - {message}")

    """ 往微信发送特殊的日志格式报告
    用法：wechat_logger.bind(wechat_user='文件传输助手').info(message)

    或者：
    # 先做好默认群名绑定
    wechat_logger = wechat_logger.bind(wechat_user='考勤管理')
    # 然后就能普通logger用法发送了
    wechat_logger.info('测试')
    """
else:
    # 降级为普通logger
    wechat_logger = logger



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2020/10/21 09:22

import os
import time
import hmac
import hashlib
import base64
import urllib.parse

import requests


class WeixinRobot:
    """ 企业微信 机器人 """

    def __init__(self, url):
        self.url = url

    def push_text(self, s):
        msgtype = 'text'
        try:
            headers = {"Content-Type": "text/plain"}
            t = {"content": s} if isinstance(s, str) else s
            data = {"msgtype": msgtype, msgtype: t}  # msgtype: text、markdown
            requests.post(url=self.url, headers=headers, json=data)
        except requests.exceptions.ConnectionError:  # 没网发送失败的时候也不报错
            pass


class DingtalkRobot:
    """ 钉钉 自定义webhook机器人

    https://ding-doc.dingtalk.com/doc#/serverapi2/qf2nxq
    """

    def __init__(self, access_token=None, secret=None):
        access_token = access_token or os.getenv('DINGTALK_ROBOT_SECRET')
        self.url = f'https://oapi.dingtalk.com/robot/send?access_token={access_token}'
        self.url += self.add_secret(secret or os.getenv('DINGTALK_ROBOT_SECRET'))
        self.headers = {"Content-Type": "application/json"}

    @classmethod
    def add_secret(cls, secret):
        """ 钉钉机器人需要加签，确保安全性 """
        timestamp = str(round(time.time() * 1000))
        secret_enc = secret.encode('utf-8')
        string_to_sign = '{}\n{}'.format(timestamp, secret)
        string_to_sign_enc = string_to_sign.encode('utf-8')
        hmac_code = hmac.new(secret_enc, string_to_sign_enc, digestmod=hashlib.sha256).digest()
        sign = urllib.parse.quote_plus(base64.b64encode(hmac_code))
        return f'&timestamp={timestamp}&sign={sign}'

    def send_data(self, data):
        try:
            requests.post(url=self.url, headers=self.headers, json=data)
        except requests.exceptions.ConnectionError as e:  # 没网发送失败的时候也不报错
            raise e

    def send_text(self, content):
        msgtype = 'text'
        d = {}
        if content: d['content'] = content
        data = {"msgtype": msgtype, msgtype: d}
        self.send_data(data)

    def send_link(self, text='', title='', pic_url='', message_url=''):
        msgtype = 'link'
        d = {}
        if text: d['text'] = text
        if title: d['title'] = title
        if pic_url: d['picUrl'] = pic_url
        if message_url: d['messageUrl'] = message_url
        data = {"msgtype": msgtype, msgtype: d}
        self.send_data(data)

    def send_markdown(self, text='', title=''):
        msgtype = 'link'
        d = {}
        if text: d['text'] = text
        if title: d['title'] = title
        data = {"msgtype": msgtype, msgtype: d}
        self.send_data(data)

    def send_actioncard(self, text='', title='', siggle_url='', siggle_title='', btn_orientation='0'):
        raise NotImplementedError

    def send_feedcard(self):
        raise NotImplementedError


class DingtalkRobot2(DingtalkRobot):

    def __init__(self, title=None):
        super().__init__()
        self.title = title

    def send_text2(self, text):  # 增加一个更加定制化的便捷接口
        from pyxllib.prog.pupil import utc_timestamp

        if self.title:
            self.send_text(f'{utc_timestamp()} {get_host_nickname()} [{self.title}] {text}')
        else:
            self.send_text(f'{utc_timestamp()} {get_host_nickname()} {text}')

    def __enter__(self):
        self.send_text2('启动')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        from pyxllib.prog.pupil import format_exception

        if exc_tb is None:
            self.send_text2('完成')
        else:
            self.send_text2(f'报错\n{format_exception(exc_val, 3)}')

if __name__ == '__main__':
    pass

    from pyxllib.prog.newbie import typename



# This is where we handle translating css styles into openpyxl styles
# and cascading those from parent to child in the dom.

from openpyxl.cell import cell
from openpyxl.styles import Font, Alignment, PatternFill, NamedStyle, Border, Side, Color
from openpyxl.styles.fills import FILL_SOLID
from openpyxl.styles.numbers import FORMAT_CURRENCY_USD_SIMPLE, FORMAT_PERCENTAGE
from openpyxl.styles.colors import BLACK

FORMAT_DATE_MMDDYYYY = 'mm/dd/yyyy'


def colormap(color):
    """
    Convenience for looking up known colors
    """
    cmap = {'black': BLACK}
    return cmap.get(color, color)


def style_string_to_dict(style):
    """
    Convert css style string to a python dictionary
    """
    def clean_split(string, delim):
        return (s.strip() for s in string.split(delim))
    styles = [clean_split(s, ":") for s in style.split(";") if ":" in s]
    return dict(styles)


def get_side(style, name):
    return {'border_style': style.get('border-{}-style'.format(name)),
            'color': colormap(style.get('border-{}-color'.format(name)))}

known_styles = {}


def style_dict_to_named_style(style_dict, number_format=None):
    """
    Change css style (stored in a python dictionary) to openpyxl NamedStyle
    """

    style_and_format_string = str({
        'style_dict': style_dict,
        'parent': style_dict.parent,
        'number_format': number_format,
    })

    if style_and_format_string not in known_styles:
        # Font
        font = Font(bold=style_dict.get('font-weight') == 'bold',
                    color=style_dict.get_color('color', None),
                    size=style_dict.get('font-size'))

        # Alignment
        alignment = Alignment(horizontal=style_dict.get('text-align', 'general'),
                              vertical=style_dict.get('vertical-align'),
                              wrap_text=style_dict.get('white-space', 'nowrap') == 'normal')

        # Fill
        bg_color = style_dict.get_color('background-color')
        fg_color = style_dict.get_color('foreground-color', Color())
        fill_type = style_dict.get('fill-type')
        if bg_color and bg_color != 'transparent':
            fill = PatternFill(fill_type=fill_type or FILL_SOLID,
                               start_color=bg_color,
                               end_color=fg_color)
        else:
            fill = PatternFill()

        # Border
        border = Border(left=Side(**get_side(style_dict, 'left')),
                        right=Side(**get_side(style_dict, 'right')),
                        top=Side(**get_side(style_dict, 'top')),
                        bottom=Side(**get_side(style_dict, 'bottom')),
                        diagonal=Side(**get_side(style_dict, 'diagonal')),
                        diagonal_direction=None,
                        outline=Side(**get_side(style_dict, 'outline')),
                        vertical=None,
                        horizontal=None)

        name = 'Style {}'.format(len(known_styles) + 1)

        pyxl_style = NamedStyle(name=name, font=font, fill=fill, alignment=alignment, border=border,
                                number_format=number_format)

        known_styles[style_and_format_string] = pyxl_style

    return known_styles[style_and_format_string]


class StyleDict(dict):
    """
    It's like a dictionary, but it looks for items in the parent dictionary
    """
    def __init__(self, *args, **kwargs):
        self.parent = kwargs.pop('parent', None)
        super(StyleDict, self).__init__(*args, **kwargs)

    def __getitem__(self, item):
        if item in self:
            return super(StyleDict, self).__getitem__(item)
        elif self.parent:
            return self.parent[item]
        else:
            raise KeyError('{} not found'.format(item))

    def __hash__(self):
        return hash(tuple([(k, self.get(k)) for k in self._keys()]))

    # Yielding the keys avoids creating unnecessary data structures
    # and happily works with both python2 and python3 where the
    # .keys() method is a dictionary_view in python3 and a list in python2.
    def _keys(self):
        yielded = set()
        for k in self.keys():
            yielded.add(k)
            yield k
        if self.parent:
            for k in self.parent._keys():
                if k not in yielded:
                    yielded.add(k)
                    yield k

    def get(self, k, d=None):
        try:
            return self[k]
        except KeyError:
            return d

    def get_color(self, k, d=None):
        """
        Strip leading # off colors if necessary
        """
        color = self.get(k, d)
        if hasattr(color, 'startswith') and color.startswith('#'):
            color = color[1:]
            if len(color) == 3:  # Premailers reduces colors like #00ff00 to #0f0, openpyxl doesn't like that
                color = ''.join(2 * c for c in color)
        return color


class Element(object):
    """
    Our base class for representing an html element along with a cascading style.
    The element is created along with a parent so that the StyleDict that we store
    can point to the parent's StyleDict.
    """
    def __init__(self, element, parent=None):
        self.element = element
        self.number_format = None
        parent_style = parent.style_dict if parent else None
        self.style_dict = StyleDict(style_string_to_dict(element.get('style', '')), parent=parent_style)
        self._style_cache = None

    def style(self):
        """
        Turn the css styles for this element into an openpyxl NamedStyle.
        """
        if not self._style_cache:
            self._style_cache = style_dict_to_named_style(self.style_dict, number_format=self.number_format)
        return self._style_cache

    def get_dimension(self, dimension_key):
        """
        Extracts the dimension from the style dict of the Element and returns it as a float.
        """
        dimension = self.style_dict.get(dimension_key)
        if dimension:
            if dimension[-2:] in ['px', 'em', 'pt', 'in', 'cm']:
                dimension = dimension[:-2]
            dimension = float(dimension)
        return dimension


class Table(Element):
    """
    The concrete implementations of Elements are semantically named for the types of elements we are interested in.
    This defines a very concrete tree structure for html tables that we expect to deal with. I prefer this compared to
    allowing Element to have an arbitrary number of children and dealing with an abstract element tree.
    """
    def __init__(self, table):
        """
        takes an html table object (from lxml)
        """
        super(Table, self).__init__(table)
        table_head = table.find('thead')
        self.head = TableHead(table_head, parent=self) if table_head is not None else None
        table_body = table.find('tbody')
        self.body = TableBody(table_body if table_body is not None else table, parent=self)


class TableHead(Element):
    """
    This class maps to the `<th>` element of the html table.
    """
    def __init__(self, head, parent=None):
        super(TableHead, self).__init__(head, parent=parent)
        self.rows = [TableRow(tr, parent=self) for tr in head.findall('tr')]


class TableBody(Element):
    """
    This class maps to the `<tbody>` element of the html table.
    """
    def __init__(self, body, parent=None):
        super(TableBody, self).__init__(body, parent=parent)
        self.rows = [TableRow(tr, parent=self) for tr in body.findall('tr')]


class TableRow(Element):
    """
    This class maps to the `<tr>` element of the html table.
    """
    def __init__(self, tr, parent=None):
        super(TableRow, self).__init__(tr, parent=parent)
        self.cells = [TableCell(cell, parent=self) for cell in tr.findall('th') + tr.findall('td')]


def element_to_string(el):
    return _element_to_string(el).strip()


def _element_to_string(el):
    string = ''

    for x in el.iterchildren():
        string += '\n' + _element_to_string(x)

    text = el.text.strip() if el.text else ''
    tail = el.tail.strip() if el.tail else ''

    return text + string + '\n' + tail


class TableCell(Element):
    """
    This class maps to the `<td>` element of the html table.
    """
    CELL_TYPES = {'TYPE_STRING', 'TYPE_FORMULA', 'TYPE_NUMERIC', 'TYPE_BOOL', 'TYPE_CURRENCY', 'TYPE_PERCENTAGE',
                  'TYPE_NULL', 'TYPE_INLINE', 'TYPE_ERROR', 'TYPE_FORMULA_CACHE_STRING', 'TYPE_INTEGER'}

    def __init__(self, cell, parent=None):
        super(TableCell, self).__init__(cell, parent=parent)
        self.value = element_to_string(cell)
        self.number_format = self.get_number_format()

    def data_type(self):
        cell_types = self.CELL_TYPES & set(self.element.get('class', '').split())
        if cell_types:
            if 'TYPE_FORMULA' in cell_types:
                # Make sure TYPE_FORMULA takes precedence over the other classes in the set.
                cell_type = 'TYPE_FORMULA'
            elif cell_types & {'TYPE_CURRENCY', 'TYPE_INTEGER', 'TYPE_PERCENTAGE'}:
                cell_type = 'TYPE_NUMERIC'
            else:
                cell_type = cell_types.pop()
        else:
            cell_type = 'TYPE_STRING'
        return getattr(cell, cell_type)

    def get_number_format(self):
        if 'TYPE_CURRENCY' in self.element.get('class', '').split():
            return FORMAT_CURRENCY_USD_SIMPLE
        if 'TYPE_INTEGER' in self.element.get('class', '').split():
            return '#,##0'
        if 'TYPE_PERCENTAGE' in self.element.get('class', '').split():
            return FORMAT_PERCENTAGE
        if 'TYPE_DATE' in self.element.get('class', '').split():
            return FORMAT_DATE_MMDDYYYY
        if self.data_type() == cell.TYPE_NUMERIC:
            try:
                int(self.value)
            except ValueError:
                return '#,##0.##'
            else:
                return '#,##0'

    def format(self, cell):
        cell.style = self.style()
        data_type = self.data_type()
        if data_type:
            cell.data_type = data_type


# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import sys
__dir__ = os.path.dirname(os.path.abspath(__file__))
sys.path.append(__dir__)
sys.path.append(os.path.abspath(os.path.join(__dir__, '../..')))

import cv2
import json
from tqdm import tqdm
from pyxlpr.ppstructure.table.table_metric import TEDS
from pyxlpr.ppstructure.table.predict_table import TableSystem
from pyxlpr.ppstructure.utility import init_args
from pyxlpr.ppocr.utils.logging import get_logger

logger = get_logger()


def parse_args():
    parser = init_args()
    parser.add_argument("--gt_path", type=str)
    return parser.parse_args()

def main(gt_path, img_root, args):
    teds = TEDS(n_jobs=16)

    text_sys = TableSystem(args)
    jsons_gt = json.load(open(gt_path))  # gt
    pred_htmls = []
    gt_htmls = []
    for img_name in tqdm(jsons_gt):
        # read image
        img = cv2.imread(os.path.join(img_root,img_name))
        pred_html = text_sys(img)
        pred_htmls.append(pred_html)

        gt_structures, gt_bboxes, gt_contents = jsons_gt[img_name]
        gt_html, gt = get_gt_html(gt_structures, gt_contents)
        gt_htmls.append(gt_html)
    scores = teds.batch_evaluate_html(gt_htmls, pred_htmls)
    logger.info('teds:', sum(scores) / len(scores))


def get_gt_html(gt_structures, gt_contents):
    end_html = []
    td_index = 0
    for tag in gt_structures:
        if '</td>' in tag:
            if gt_contents[td_index] != []:
                end_html.extend(gt_contents[td_index])
            end_html.append(tag)
            td_index += 1
        else:
            end_html.append(tag)
    return ''.join(end_html), end_html


if __name__ == '__main__':
    args = parse_args()
    main(args.gt_path,args.image_dir, args)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/06/03 22:08

""" pyxllib常用功能
"""

try:
    import fire
except:
    # 除了模块找不到，在labelme场景可能还有AttributeError: 'NoneType' object has no attribute 'isatty'
    # 遇到这种情况，都可以忽略
    pass

from pyxllib.file.packlib import *

from pyxllib.prog.newbie import *
from pyxllib.prog.pupil import *
from pyxllib.prog.specialist import *
# from pyxllib.prog.deprecatedlib import deprecated
from deprecated import deprecated

from pyxllib.algo.newbie import *
from pyxllib.algo.pupil import *
from pyxllib.algo.specialist import *

from pyxllib.text.newbie import *
from pyxllib.text.pupil import *
from pyxllib.text.specialist import *

from pyxllib.file.newbie import *
from pyxllib.file.pupil import *
from pyxllib.file.specialist import *

if __name__ == '__main__':
    # 直接运行的话，支持开放出所有函数类接口
    fire.Fire()


    from pyxllib.ext.demolib import test_re


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2024/11/13

import os
import tempfile
import time

from filelock import FileLock, Timeout


class XlFileLock(FileLock):
    def __init__(self, lock_file: str, timeout: float = -1, **kwargs):
        """
        初始化 XlFileLock 实例，锁文件默认存储在系统临时目录中。

        :param str filename: 锁文件的文件名（不包含路径）。
        :param float timeout: 获取锁的超时时间（以秒为单位），默认值为 -1 表示无限等待。
        """
        lock_file_path = os.path.join(tempfile.gettempdir(), lock_file)
        super().__init__(lock_file=lock_file_path, timeout=timeout, **kwargs)


def get_autogui_lock(lock_file='autogui.lock', timeout=-1):
    """
    注意：XlFileLock中的lock_file无法设置默认值，这跟FileLock的底层有关
        所以无法继承一个AutoUiLock类，但是可以通过这种函数的方式，绕开它特殊的初始化限制
    """
    lock = XlFileLock(lock_file=lock_file, timeout=timeout)
    return lock


if __name__ == "__main__":
    try:
        with get_autogui_lock(timeout=5):
            print("Lock acquired, doing some work...")
            time.sleep(3)
            print("Work done.")
    except Timeout:
        print("Failed to acquire lock within the timeout.")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/06/06 10:51


def typename(c):
    """ 简化输出的type类型

    >>> typename(123)
    'int'
    """
    return str(type(c))[8:-2]


class SingletonForEveryClass(type):
    """ 注意如果A是单例类，B从A继承，那么实际有且仅有A、B两个不同的实例对象 """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        tag = f'{cls}'
        # 其实转字符串来判断是不太严谨的，有些类型字符串后的显示效果是一模一样的
        # dprint(tag)
        if tag not in cls._instances:
            cls._instances[tag] = super(SingletonForEveryClass, cls).__call__(*args, **kwargs)
        return cls._instances[tag]


class SingletonForEveryInitArgs(type):
    """ Python单例模式(Singleton)的N种实现 - 知乎: https://zhuanlan.zhihu.com/p/37534850

    注意！注意！注意！重要的事说三遍！
    这里的单例类不是传统意义上的单例类。
    传统意义的单例类，不管用怎样不同的初始化参数创建对象，永远都只有最初的那个对象类。
    但是这个单例类，为每种不同的参数创建形式，都构造了一个对象。
    """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        tag = f'{cls}{args}{kwargs}'  # id加上所有参数的影响来控制单例类
        # 其实转字符串来判断是不太严谨的，有些类型字符串后的显示效果是一模一样的
        # dprint(tag)
        if tag not in cls._instances:
            cls._instances[tag] = super(SingletonForEveryInitArgs, cls).__call__(*args, **kwargs)
        return cls._instances[tag]


def xlbool(v):
    """ 有些类型不能直接判断真假，例如具有多值的np.array，df等

    这些有歧义的情况，在我的mybool里暂时都判断为True，如果有需要，需要精细化判断，可以扩展自己的npbool、dfbool
    """
    try:
        return bool(v)
    except ValueError:
        return True


def document(func):
    """文档函数装饰器
    用该装饰器器时，表明一个函数是用伪代码在表示一系列的操作逻辑，不能直接拿来执行的
    很可能是一套半自动化工具
    """

    def wrapper(*args):
        raise RuntimeError(f'函数:{func.__name__} 是一个伪代码流程示例文档，不能直接运行')

    return wrapper


class RunOnlyOnce:
    """ 被装饰的函数，不同的参数输入形式，只会被执行一次，

    重复执行时会从内存直接调用上次相同参数调用下的运行的结果
    可以使用reset成员函数重置，下一次调用该函数时则会重新执行

    文档：https://www.yuque.com/xlpr/pyxllib/RunOnlyOnce

    使用好该装饰器，可能让一些动态规划dp、搜索问题变得更简洁，
    以及一些配置文件操作，可以做到只读一遍
    """

    def __init__(self, func, distinct_args=True):
        """
        :param func: 封装的函数
        :param distinct_args: 默认不同输入参数形式，都会保存一个结果
            设为False，则不管何种参数形式，函数就真的只会保存第一次运行的结果
        """
        self.func = func
        self.distinct_args = distinct_args
        self.results = {}

    @classmethod
    def decorator(cls, distinct_args=True):
        """ 作为装饰器的时候，如果要设置参数，要用这个接口 """

        def wrap(func):
            return cls(func, distinct_args)

        return wrap

    def __call__(self, *args, **kwargs):
        tag = f'{args}{kwargs}' if self.distinct_args else ''
        # TODO 思考更严谨，考虑了值类型的tag标记
        #   目前的tag规则，是必要不充分条件。还可以使用id，则是充分不必要条件
        #   能找到充要条件来做是最好的，不行的话，也应该用更严谨的充分条件来做
        # TODO kwargs的顺序应该是没影响的，要去掉顺序干扰
        if tag not in self.results:
            self.results[tag] = self.func(*args, **kwargs)
        return self.results[tag]

    def reset(self):
        self.results = {}


def len_in_dim2(arr):
    """ 计算类List结构在第2维上的最大长度

    >>> len_in_dim2([[1,1], [2], [3,3,3]])
    3

    >>> len_in_dim2([1, 2, 3])  # TODO 是不是应该改成0合理？但不知道牵涉到哪些功能影响
    1
    """
    if not isinstance(arr, (list, tuple)):
        raise TypeError('类型错误，不是list构成的二维数组')

    # 找出元素最多的列
    column_num = 0
    for i, item in enumerate(arr):
        if isinstance(item, (list, tuple)):  # 该行是一个一维数组
            column_num = max(column_num, len(item))
        else:  # 如果不是数组，是指单个元素，当成1列处理
            column_num = max(column_num, 1)

    return column_num


def ensure_array(arr, default_value=''):
    """对一个由list、tuple组成的二维数组，确保所有第二维的列数都相同

    >>> ensure_array([[1,1], [2], [3,3,3]])
    [[1, 1, ''], [2, '', ''], [3, 3, 3]]
    """
    max_cols = len_in_dim2(arr)
    if max_cols == 1:
        return arr
    dv = str(default_value)
    a = [[]] * len(arr)
    for i, ls in enumerate(arr):
        if isinstance(ls, (list, tuple)):
            t = list(arr[i])
        else:
            t = [ls]  # 如果不是数组，是指单个元素，当成1列处理
        a[i] = t + [dv] * (max_cols - len(t))  # 左边的写list，是防止有的情况是tuple，要强制转list后拼接
    return a


def swap_rowcol(a, *, ensure_arr=False, default_value=''):
    """矩阵行列互换

    注：如果列数是不均匀的，则会以最小列数作为行数

    >>> swap_rowcol([[1,2,3], [4,5,6]])
    [[1, 4], [2, 5], [3, 6]]
    """
    if ensure_arr:
        a = ensure_array(a, default_value)
    # 这是非常有教学意义的行列互换实现代码
    return list(map(list, zip(*a)))


class GrowingList(list):
    """可变长list"""

    def __init__(self, default_value=None):
        super().__init__(self)
        self.default_value = default_value

    def __getitem__(self, index):
        if index >= len(self):
            self.extend([self.default_value] * (index + 1 - len(self)))
        return list.__getitem__(self, index)

    def __setitem__(self, index, value):
        if index >= len(self):
            self.extend([self.default_value] * (index + 1 - len(self)))
        list.__setitem__(self, index, value)


class GenFunction:
    """ 一般用来生成高阶函数的函数对象

    这个名字可能还不是很精确，后面有想法再改
    """

    @classmethod
    def ensure_func(cls, x, default):
        """ 确保x是callable对象，否则用default初始化 """
        if callable(x):
            return x
        else:
            return default


def first_nonnone(args, judge=None):
    """ 返回第1个满足条件的值

    :param args: 参数清单
    :param judge: 判断器，默认返回第一个非None值，也可以自定义判定函数
    """
    judge = GenFunction.ensure_func(judge, lambda x: x is not None)
    for x in args:
        if judge(x):
            return x
    return args[-1]  # 全部都不满足，返回最后一个值


def round_int(x, *, ndim=0):
    """ 先四舍五入，再取整

    :param x: 一个数值，或者多维数组
    :param ndim: x是数值是默认0，否则指定数组维度，批量处理
        比如ndim=1是一维数组
        ndim=2是二维数组

    >>> round_int(1.5)
    2
    >>> round_int(1.4)
    1
    >>> round_int([2.3, 1.42], ndim=1)
    [2, 1]
    >>> round_int([[2.3, 1.42], [3.6]], ndim=2)
    [[2, 1], [4]]
    """
    if ndim:
        return [round_int(a, ndim=ndim - 1) for a in x]
    else:
        return int(round(x, 0))


def human_readable_size(n):
    """ 我个人习惯常用的size显示方式 """
    for u in [' B', 'KB', 'MB', 'GB']:
        if n < 1024:
            return f'{round_int(n)}{u}'
        else:
            n /= 1024
    else:
        return f'{round_int(n)}TB'


def human_readable_number(value, base_type='K', precision=4):
    """ 数字美化输出函数

    :param float|int value: 要转换的数值
    :param int precision: 有效数字的长度
    :param str base_type: 进制类型，'K'为1000进制, 'KB'为1024进制（KiB同理）, '万'为中文万进制
    :return: 美化后的字符串
    """
    if value is None:
        return ''

    if abs(value) < 1:
        return f'{value:.{precision}g}'

    # 设置不同进制的单位和基数
    units, base = {
        'K': (['', 'K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y'], 1000),
        'KB': (['', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB'], 1024),
        'KiB': (['', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB', 'ZiB', 'YiB'], 1024),
        '万': (['', '万', '亿', '万亿', '亿亿'], 10000),
        '秒': (['秒', [60, '分'], [60, '时'], [24, '天'], [7, '周'], [4.345, '月'], [12, '年']], 60),
    }.get(base_type, ([''], 1))  # 默认为空单位和基数1

    x, i = abs(value), 0
    while x >= base and i < len(units) - 1:
        x /= base
        i += 1
        if isinstance(units[i], list):
            base = units[i][0]
            units[i] = units[i][1]

    x = f'{x:.{precision}g}'  # 四舍五入到指定精度
    prefix = '-' if value < 0 else ''  # 负数处理
    return f"{prefix}{x}{units[i]}"


class CvtType:
    """ 这些系列的转换函数，转换失败统一报错为ValueError

    返回异常会比较好，否则返回None、False之类的，
        有时候可能就是要转换的值呢，会有歧义
    """

    @classmethod
    def str2list(cls, arg):
        try:
            res = eval(arg)
        except SyntaxError:
            raise ValueError

        if not isinstance(res, list):
            raise ValueError

        return res

    @classmethod
    def str2dict(cls, arg):
        try:
            res = eval(arg)
        except SyntaxError:
            raise ValueError

        if not isinstance(res, dict):
            raise ValueError

        return res

    @classmethod
    def factory(cls, name):
        """ 从字符串名称，映射到对应的转换函数 """
        return {'int': int,
                'float': float,
                'str': str,
                'list': cls.str2list,
                'dict': cls.str2dict}.get(name, None)


def mod_minabs(x, m):
    a = x % m
    return a if a < m / 2 else a - m


class classproperty(property):
    """ python - Using property() on classmethods - Stack Overflow
        https://stackoverflow.com/questions/128573/using-property-on-classmethods
    """

    def __get__(self, obj, objtype=None):
        return super(classproperty, self).__get__(objtype)

    def __set__(self, obj, value):
        super(classproperty, self).__set__(type(obj), value)

    def __delete__(self, obj):
        super(classproperty, self).__delete__(type(obj))


def decode_bitflags(n, flags, return_type=dict):
    """ 解析一个位标记的功能

    :param int n: 一个整数标记
    :param list|tuple flags: 一个对应明文数组
        flags[0]对应2**0的明文，flags[1]对应2**1的明文，以此类推
    :param type return_type: 返回的数据类型
        默认dict，记录所有key的bool结果
        可以设set，只返回真的标记结果

    >>> decode_bitflags(18, ('superscript', 'italic', 'serifed', 'monospaced', 'bold'))
    {'superscript': 2, 'italic': 0, 'serifed': 0, 'monospaced': 16, 'bold': 0}
    >>> decode_bitflags(18, ('superscript', 'italic', 'serifed', 'monospaced', 'bold'), set)
    {'superscript', 'monospaced'}
    """
    if return_type == dict:
        return {x: n & (2 << i) for i, x in enumerate(flags)}
    elif return_type == set:
        return {x for i, x in enumerate(flags) if (n & (2 << i))}
    else:
        raise ValueError


def xl_format_g(x, p=3):
    """ 普通format的g模式不太满足我的需求，做了点修改

    注：g是比较方便的一种数值格式化方法，会比较智能地判断是否整数显示，或者普通显示、科学计数法显示

    :param x: 数值x
    """
    s = f'{x:g}'
    if 'e' in s:
        # 如果变成了科学计数法，明确保留3位有效数字
        return '{:.{}g}'.format(x, p=3)
    else:
        # 否则返回默认的g格式
        return s


class EmptyWith:
    """ 空上下文类
    """

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


def funcmsg(func):
    """输出函数func所在的文件、函数名、函数起始行"""
    # showdir(func)
    if not hasattr(func, '__name__'):  # 没有__name__属性表示这很可能是一个装饰器去处理原函数了
        if hasattr(func, 'func'):  # 我的装饰器常用func成员存储原函数对象
            func = func.func
        else:
            return f'装饰器：{type(func)}，无法定位'
    return f'函数名：{func.__name__}，来自文件：{func.__code__.co_filename}，所在行号={func.__code__.co_firstlineno}'


def set_global_var(name, value):
    """ 设置某个全局变量，一般用在一些特殊的需要跨作用域进行调试的场景
    切勿！切勿！切勿用于正式功能，否则会导致难以维护控制的功能代码

    为了避免和某些关键的全局变量名冲突，这里的变量命令统一会加上pyxllib_的前缀
    """
    g = globals()
    name = f'pyxllib_{name}'
    g[name] = value


def get_global_var(name, default_value=None):
    """ 获取某个全局变量的值 """
    g = globals()
    name = f'pyxllib_{name}'
    if name not in g:
        g[name] = default_value
    return g[name]


def convert_to_json_compatible(d, custom_converter=None):
    """ 递归地将字典等类型转换为JSON兼容格式。对于非标准JSON类型，使用自定义转换器或默认转换为字符串。

    :param d: 要转换的字典或列表。
    :param custom_converter: 自定义转换函数，用于处理非标准JSON类型的值。
    :return: 转换后的字典或列表。

    todo 这个函数想法是好的，但总感觉精确性中，总容易有些问题的，需要更多的考察测试
    """
    if custom_converter is None:
        custom_converter = str

    if isinstance(d, dict):  # defaultdict呢？
        return {k: convert_to_json_compatible(v, custom_converter) for k, v in d.items()}
    elif isinstance(d, list):
        return [convert_to_json_compatible(v, custom_converter) for v in d]
    elif isinstance(d, (int, float, str, bool)) or d is None:
        return d
    else:
        return custom_converter(d)



# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import sys
import cv2
import numpy as np
import paddle
from PIL import Image, ImageDraw, ImageFont
import math
from paddle import inference
import time
from pyxlpr.ppocr.utils.logging import get_logger

from pyxllib.xl import get_font_file

def str2bool(v):
    return v.lower() in ("true", "t", "1")


def init_args():
    parser = argparse.ArgumentParser()
    # params for prediction engine
    parser.add_argument("--use_gpu", type=str2bool, default=True)
    parser.add_argument("--ir_optim", type=str2bool, default=True)
    parser.add_argument("--use_tensorrt", type=str2bool, default=False)
    parser.add_argument("--min_subgraph_size", type=int, default=15)
    parser.add_argument("--precision", type=str, default="fp32")
    parser.add_argument("--gpu_mem", type=int, default=500)

    # params for text detector
    parser.add_argument("--image_dir", type=str)
    parser.add_argument("--det_algorithm", type=str, default='DB')
    parser.add_argument("--det_model_dir", type=str)
    parser.add_argument("--det_limit_side_len", type=float, default=960)
    parser.add_argument("--det_limit_type", type=str, default='max')

    # DB parmas
    parser.add_argument("--det_db_thresh", type=float, default=0.3)
    parser.add_argument("--det_db_box_thresh", type=float, default=0.6)
    parser.add_argument("--det_db_unclip_ratio", type=float, default=1.5)
    parser.add_argument("--max_batch_size", type=int, default=10)
    parser.add_argument("--use_dilation", type=str2bool, default=False)
    parser.add_argument("--det_db_score_mode", type=str, default="fast")
    # EAST parmas
    parser.add_argument("--det_east_score_thresh", type=float, default=0.8)
    parser.add_argument("--det_east_cover_thresh", type=float, default=0.1)
    parser.add_argument("--det_east_nms_thresh", type=float, default=0.2)

    # SAST parmas
    parser.add_argument("--det_sast_score_thresh", type=float, default=0.5)
    parser.add_argument("--det_sast_nms_thresh", type=float, default=0.2)
    parser.add_argument("--det_sast_polygon", type=str2bool, default=False)

    # PSE parmas
    parser.add_argument("--det_pse_thresh", type=float, default=0)
    parser.add_argument("--det_pse_box_thresh", type=float, default=0.85)
    parser.add_argument("--det_pse_min_area", type=float, default=16)
    parser.add_argument("--det_pse_box_type", type=str, default='box')
    parser.add_argument("--det_pse_scale", type=int, default=1)

    # params for text recognizer
    parser.add_argument("--rec_algorithm", type=str, default='CRNN')
    parser.add_argument("--rec_model_dir", type=str)
    parser.add_argument("--rec_image_shape", type=str, default="3, 32, 320")
    parser.add_argument("--rec_batch_num", type=int, default=6)
    parser.add_argument("--max_text_length", type=int, default=25)
    parser.add_argument(
        "--rec_char_dict_path",
        type=str,
        default="./ppocr/utils/ppocr_keys_v1.txt")
    parser.add_argument("--use_space_char", type=str2bool, default=True)
    parser.add_argument(
        "--vis_font_path", type=str, default="./doc/fonts/simfang.ttf")
    parser.add_argument("--drop_score", type=float, default=0.5)

    # params for e2e
    parser.add_argument("--e2e_algorithm", type=str, default='PGNet')
    parser.add_argument("--e2e_model_dir", type=str)
    parser.add_argument("--e2e_limit_side_len", type=float, default=768)
    parser.add_argument("--e2e_limit_type", type=str, default='max')

    # PGNet parmas
    parser.add_argument("--e2e_pgnet_score_thresh", type=float, default=0.5)
    parser.add_argument(
        "--e2e_char_dict_path", type=str, default="./ppocr/utils/ic15_dict.txt")
    parser.add_argument("--e2e_pgnet_valid_set", type=str, default='totaltext')
    parser.add_argument("--e2e_pgnet_mode", type=str, default='fast')

    # params for text classifier
    parser.add_argument("--use_angle_cls", type=str2bool, default=False)
    parser.add_argument("--cls_model_dir", type=str)
    parser.add_argument("--cls_image_shape", type=str, default="3, 48, 192")
    parser.add_argument("--label_list", type=list, default=['0', '180'])
    parser.add_argument("--cls_batch_num", type=int, default=6)
    parser.add_argument("--cls_thresh", type=float, default=0.9)

    parser.add_argument("--enable_mkldnn", type=str2bool, default=False)
    parser.add_argument("--cpu_threads", type=int, default=10)
    parser.add_argument("--use_pdserving", type=str2bool, default=False)
    parser.add_argument("--warmup", type=str2bool, default=False)

    #
    parser.add_argument(
        "--draw_img_save_dir", type=str, default="./inference_results")
    parser.add_argument("--save_crop_res", type=str2bool, default=False)
    parser.add_argument("--crop_res_save_dir", type=str, default="./output")

    # multi-process
    parser.add_argument("--use_mp", type=str2bool, default=False)
    parser.add_argument("--total_process_num", type=int, default=1)
    parser.add_argument("--process_id", type=int, default=0)

    parser.add_argument("--benchmark", type=str2bool, default=False)
    parser.add_argument("--save_log_path", type=str, default="./log_output/")

    parser.add_argument("--show_log", type=str2bool, default=True)
    parser.add_argument("--use_onnx", type=str2bool, default=False)
    return parser


def parse_args():
    parser = init_args()
    return parser.parse_args()


def create_predictor(args, mode, logger):
    if mode == "det":
        model_dir = args.det_model_dir
    elif mode == 'cls':
        model_dir = args.cls_model_dir
    elif mode == 'rec':
        model_dir = args.rec_model_dir
    elif mode == 'table':
        model_dir = args.table_model_dir
    else:
        model_dir = args.e2e_model_dir

    if model_dir is None:
        logger.info("not find {} model file path {}".format(mode, model_dir))
        sys.exit(0)
    if args.use_onnx:
        import onnxruntime as ort
        model_file_path = model_dir
        if not os.path.exists(model_file_path):
            raise ValueError("not find model file path {}".format(
                model_file_path))
        sess = ort.InferenceSession(model_file_path)
        return sess, sess.get_inputs()[0], None, None

    else:
        model_file_path = model_dir + "/inference.pdmodel"
        params_file_path = model_dir + "/inference.pdiparams"
        if not os.path.exists(model_file_path):
            raise ValueError("not find model file path {}".format(
                model_file_path))
        if not os.path.exists(params_file_path):
            raise ValueError("not find params file path {}".format(
                params_file_path))

        config = inference.Config(model_file_path, params_file_path)

        if hasattr(args, 'precision'):
            if args.precision == "fp16" and args.use_tensorrt:
                precision = inference.PrecisionType.Half
            elif args.precision == "int8":
                precision = inference.PrecisionType.Int8
            else:
                precision = inference.PrecisionType.Float32
        else:
            precision = inference.PrecisionType.Float32

        if args.use_gpu:
            gpu_id = get_infer_gpuid()
            if gpu_id is None:
                logger.warning(
                    "GPU is not found in current device by nvidia-smi. Please check your device or ignore it if run on jeston."
                )
            config.enable_use_gpu(args.gpu_mem, 0)
            if args.use_tensorrt:
                config.enable_tensorrt_engine(
                    workspace_size=1 << 30,
                    precision_mode=precision,
                    max_batch_size=args.max_batch_size,
                    min_subgraph_size=args.min_subgraph_size)
                # skip the minmum trt subgraph
            use_dynamic_shape = True
            if mode == "det":
                min_input_shape = {
                    "x": [1, 3, 50, 50],
                    "conv2d_92.tmp_0": [1, 120, 20, 20],
                    "conv2d_91.tmp_0": [1, 24, 10, 10],
                    "conv2d_59.tmp_0": [1, 96, 20, 20],
                    "nearest_interp_v2_1.tmp_0": [1, 256, 10, 10],
                    "nearest_interp_v2_2.tmp_0": [1, 256, 20, 20],
                    "conv2d_124.tmp_0": [1, 256, 20, 20],
                    "nearest_interp_v2_3.tmp_0": [1, 64, 20, 20],
                    "nearest_interp_v2_4.tmp_0": [1, 64, 20, 20],
                    "nearest_interp_v2_5.tmp_0": [1, 64, 20, 20],
                    "elementwise_add_7": [1, 56, 2, 2],
                    "nearest_interp_v2_0.tmp_0": [1, 256, 2, 2]
                }
                max_input_shape = {
                    "x": [1, 3, 1536, 1536],
                    "conv2d_92.tmp_0": [1, 120, 400, 400],
                    "conv2d_91.tmp_0": [1, 24, 200, 200],
                    "conv2d_59.tmp_0": [1, 96, 400, 400],
                    "nearest_interp_v2_1.tmp_0": [1, 256, 200, 200],
                    "conv2d_124.tmp_0": [1, 256, 400, 400],
                    "nearest_interp_v2_2.tmp_0": [1, 256, 400, 400],
                    "nearest_interp_v2_3.tmp_0": [1, 64, 400, 400],
                    "nearest_interp_v2_4.tmp_0": [1, 64, 400, 400],
                    "nearest_interp_v2_5.tmp_0": [1, 64, 400, 400],
                    "elementwise_add_7": [1, 56, 400, 400],
                    "nearest_interp_v2_0.tmp_0": [1, 256, 400, 400]
                }
                opt_input_shape = {
                    "x": [1, 3, 640, 640],
                    "conv2d_92.tmp_0": [1, 120, 160, 160],
                    "conv2d_91.tmp_0": [1, 24, 80, 80],
                    "conv2d_59.tmp_0": [1, 96, 160, 160],
                    "nearest_interp_v2_1.tmp_0": [1, 256, 80, 80],
                    "nearest_interp_v2_2.tmp_0": [1, 256, 160, 160],
                    "conv2d_124.tmp_0": [1, 256, 160, 160],
                    "nearest_interp_v2_3.tmp_0": [1, 64, 160, 160],
                    "nearest_interp_v2_4.tmp_0": [1, 64, 160, 160],
                    "nearest_interp_v2_5.tmp_0": [1, 64, 160, 160],
                    "elementwise_add_7": [1, 56, 40, 40],
                    "nearest_interp_v2_0.tmp_0": [1, 256, 40, 40]
                }
                min_pact_shape = {
                    "nearest_interp_v2_26.tmp_0": [1, 256, 20, 20],
                    "nearest_interp_v2_27.tmp_0": [1, 64, 20, 20],
                    "nearest_interp_v2_28.tmp_0": [1, 64, 20, 20],
                    "nearest_interp_v2_29.tmp_0": [1, 64, 20, 20]
                }
                max_pact_shape = {
                    "nearest_interp_v2_26.tmp_0": [1, 256, 400, 400],
                    "nearest_interp_v2_27.tmp_0": [1, 64, 400, 400],
                    "nearest_interp_v2_28.tmp_0": [1, 64, 400, 400],
                    "nearest_interp_v2_29.tmp_0": [1, 64, 400, 400]
                }
                opt_pact_shape = {
                    "nearest_interp_v2_26.tmp_0": [1, 256, 160, 160],
                    "nearest_interp_v2_27.tmp_0": [1, 64, 160, 160],
                    "nearest_interp_v2_28.tmp_0": [1, 64, 160, 160],
                    "nearest_interp_v2_29.tmp_0": [1, 64, 160, 160]
                }
                min_input_shape.update(min_pact_shape)
                max_input_shape.update(max_pact_shape)
                opt_input_shape.update(opt_pact_shape)
            elif mode == "rec":
                if args.rec_algorithm != "CRNN":
                    use_dynamic_shape = False
                min_input_shape = {"x": [1, 3, 32, 10]}
                max_input_shape = {"x": [args.rec_batch_num, 3, 32, 1536]}
                opt_input_shape = {"x": [args.rec_batch_num, 3, 32, 320]}
            elif mode == "cls":
                min_input_shape = {"x": [1, 3, 48, 10]}
                max_input_shape = {"x": [args.rec_batch_num, 3, 48, 1024]}
                opt_input_shape = {"x": [args.rec_batch_num, 3, 48, 320]}
            else:
                use_dynamic_shape = False
            if use_dynamic_shape:
                config.set_trt_dynamic_shape_info(
                    min_input_shape, max_input_shape, opt_input_shape)

        else:
            config.disable_gpu()
            if hasattr(args, "cpu_threads"):
                config.set_cpu_math_library_num_threads(args.cpu_threads)
            else:
                # default cpu threads as 10
                config.set_cpu_math_library_num_threads(10)
            if args.enable_mkldnn:
                # cache 10 different shapes for mkldnn to avoid memory leak
                config.set_mkldnn_cache_capacity(10)
                config.enable_mkldnn()
                if args.precision == "fp16":
                    config.enable_mkldnn_bfloat16()
        # enable memory optim
        config.enable_memory_optim()
        config.disable_glog_info()

        config.delete_pass("conv_transpose_eltwiseadd_bn_fuse_pass")
        if mode == 'table':
            config.delete_pass("fc_fuse_pass")  # not supported for table
        config.switch_use_feed_fetch_ops(False)
        config.switch_ir_optim(True)

        # create predictor
        predictor = inference.create_predictor(config)
        input_names = predictor.get_input_names()
        for name in input_names:
            input_tensor = predictor.get_input_handle(name)
        output_names = predictor.get_output_names()
        output_tensors = []
        for output_name in output_names:
            output_tensor = predictor.get_output_handle(output_name)
            output_tensors.append(output_tensor)
        return predictor, input_tensor, output_tensors, config


def get_infer_gpuid():
    if os.name == 'nt':
        try:
            return int(os.environ['CUDA_VISIBLE_DEVICES'].split(',')[0])
        except KeyError:
            return 0
    if not paddle.fluid.core.is_compiled_with_rocm():
        cmd = "env | grep CUDA_VISIBLE_DEVICES"
    else:
        cmd = "env | grep HIP_VISIBLE_DEVICES"
    env_cuda = os.popen(cmd).readlines()
    if len(env_cuda) == 0:
        return 0
    else:
        gpu_id = env_cuda[0].strip().split("=")[1]
        return int(gpu_id[0])


def draw_e2e_res(dt_boxes, strs, img_path):
    src_im = cv2.imread(img_path)
    for box, str in zip(dt_boxes, strs):
        box = box.astype(np.int32).reshape((-1, 1, 2))
        cv2.polylines(src_im, [box], True, color=(255, 255, 0), thickness=2)
        cv2.putText(
            src_im,
            str,
            org=(int(box[0, 0, 0]), int(box[0, 0, 1])),
            fontFace=cv2.FONT_HERSHEY_COMPLEX,
            fontScale=0.7,
            color=(0, 255, 0),
            thickness=1)
    return src_im


def draw_text_det_res(dt_boxes, img_path):
    src_im = cv2.imread(img_path)
    for box in dt_boxes:
        box = np.array(box).astype(np.int32).reshape(-1, 2)
        cv2.polylines(src_im, [box], True, color=(255, 255, 0), thickness=2)
    return src_im


def resize_img(img, input_size=600):
    """
    resize img and limit the longest side of the image to input_size
    """
    img = np.array(img)
    im_shape = img.shape
    im_size_max = np.max(im_shape[0:2])
    im_scale = float(input_size) / float(im_size_max)
    img = cv2.resize(img, None, None, fx=im_scale, fy=im_scale)
    return img


def draw_ocr(image,
             boxes,
             txts=None,
             scores=None,
             drop_score=0.5,
             font_path='simfang.ttf'):
    """
    Visualize the results of OCR detection and recognition
    args:
        image(Image|array): RGB image
        boxes(list): boxes with shape(N, 4, 2)
        txts(list): the texts
        scores(list): txxs corresponding scores
        drop_score(float): only scores greater than drop_threshold will be visualized
        font_path: the path of font which is used to draw text
    return(array):
        the visualized img
    """
    if scores is None:
        scores = [1] * len(boxes)
    box_num = len(boxes)
    for i in range(box_num):
        if scores is not None and (scores[i] < drop_score or
                                   math.isnan(scores[i])):
            continue
        box = np.reshape(np.array(boxes[i]), [-1, 1, 2]).astype(np.int64)
        image = cv2.polylines(np.array(image), [box], True, (255, 0, 0), 2)
    if txts is not None:
        img = np.array(resize_img(image, input_size=600))
        txt_img = text_visual(
            txts,
            scores,
            img_h=img.shape[0],
            img_w=600,
            threshold=drop_score,
            font_path=font_path)
        img = np.concatenate([np.array(img), np.array(txt_img)], axis=1)
        return img
    return image


def draw_ocr_box_txt(image,
                     boxes,
                     txts,
                     scores=None,
                     drop_score=0.5,
                     font_path='simfang.ttf'):
    h, w = image.height, image.width
    img_left = image.copy()
    img_right = Image.new('RGB', (w, h), (255, 255, 255))

    import random

    random.seed(0)
    draw_left = ImageDraw.Draw(img_left)
    draw_right = ImageDraw.Draw(img_right)
    for idx, (box, txt) in enumerate(zip(boxes, txts)):
        if scores is not None and scores[idx] < drop_score:
            continue
        color = (random.randint(0, 255), random.randint(0, 255),
                 random.randint(0, 255))
        draw_left.polygon(box, fill=color)
        draw_right.polygon(
            [
                box[0][0], box[0][1], box[1][0], box[1][1], box[2][0],
                box[2][1], box[3][0], box[3][1]
            ],
            outline=color)
        box_height = math.sqrt((box[0][0] - box[3][0])**2 + (box[0][1] - box[3][
            1])**2)
        box_width = math.sqrt((box[0][0] - box[1][0])**2 + (box[0][1] - box[1][
            1])**2)
        if box_height > 2 * box_width:
            font_size = max(int(box_width * 0.9), 10)
            font = ImageFont.truetype(font_path, font_size, encoding="utf-8")
            cur_y = box[0][1]
            for c in txt:
                char_size = font.getsize(c)
                draw_right.text(
                    (box[0][0] + 3, cur_y), c, fill=(0, 0, 0), font=font)
                cur_y += char_size[1]
        else:
            font_size = max(int(box_height * 0.8), 10)
            font = ImageFont.truetype(font_path, font_size, encoding="utf-8")
            draw_right.text(
                [box[0][0], box[0][1]], txt, fill=(0, 0, 0), font=font)
    img_left = Image.blend(image, img_left, 0.5)
    img_show = Image.new('RGB', (w * 2, h), (255, 255, 255))
    img_show.paste(img_left, (0, 0, w, h))
    img_show.paste(img_right, (w, 0, w * 2, h))
    return np.array(img_show)


def str_count(s):
    """
    Count the number of Chinese characters,
    a single English character and a single number
    equal to half the length of Chinese characters.
    args:
        s(string): the input of string
    return(int):
        the number of Chinese characters
    """
    import string
    count_zh = count_pu = 0
    s_len = len(s)
    en_dg_count = 0
    for c in s:
        if c in string.ascii_letters or c.isdigit() or c.isspace():
            en_dg_count += 1
        elif c.isalpha():
            count_zh += 1
        else:
            count_pu += 1
    return s_len - math.ceil(en_dg_count / 2)


def text_visual(texts,
                scores,
                img_h=400,
                img_w=600,
                threshold=0.,
                font_path='simfang.ttf'):
    """
    create new blank img and draw txt on it
    args:
        texts(list): the text will be draw
        scores(list|None): corresponding score of each txt
        img_h(int): the height of blank img
        img_w(int): the width of blank img
        font_path: the path of font which is used to draw text
    return(array):
    """
    if scores is not None:
        assert len(texts) == len(
            scores), "The number of txts and corresponding scores must match"

    def create_blank_img():
        blank_img = np.ones(shape=[img_h, img_w], dtype=np.int8) * 255
        blank_img[:, img_w - 1:] = 0
        blank_img = Image.fromarray(blank_img).convert("RGB")
        draw_txt = ImageDraw.Draw(blank_img)
        return blank_img, draw_txt

    blank_img, draw_txt = create_blank_img()

    font_size = 20
    txt_color = (0, 0, 0)

    if not os.path.isfile(font_path):
        font_path = str(get_font_file('simfang.ttf'))

    font = ImageFont.truetype(font_path, font_size, encoding="utf-8")

    gap = font_size + 5
    txt_img_list = []
    count, index = 1, 0
    for idx, txt in enumerate(texts):
        index += 1
        if scores[idx] < threshold or math.isnan(scores[idx]):
            index -= 1
            continue
        first_line = True
        while str_count(txt) >= img_w // font_size - 4:
            tmp = txt
            txt = tmp[:img_w // font_size - 4]
            if first_line:
                new_txt = str(index) + ': ' + txt
                first_line = False
            else:
                new_txt = '    ' + txt
            draw_txt.text((0, gap * count), new_txt, txt_color, font=font)
            txt = tmp[img_w // font_size - 4:]
            if count >= img_h // gap - 1:
                txt_img_list.append(np.array(blank_img))
                blank_img, draw_txt = create_blank_img()
                count = 0
            count += 1
        if first_line:
            new_txt = str(index) + ': ' + txt + '   ' + '%.3f' % (scores[idx])
        else:
            new_txt = "  " + txt + "  " + '%.3f' % (scores[idx])
        draw_txt.text((0, gap * count), new_txt, txt_color, font=font)
        # whether add new blank img or not
        if count >= img_h // gap - 1 and idx + 1 < len(texts):
            txt_img_list.append(np.array(blank_img))
            blank_img, draw_txt = create_blank_img()
            count = 0
        count += 1
    txt_img_list.append(np.array(blank_img))
    if len(txt_img_list) == 1:
        blank_img = np.array(txt_img_list[0])
    else:
        blank_img = np.concatenate(txt_img_list, axis=1)
    return np.array(blank_img)


def base64_to_cv2(b64str):
    import base64
    data = base64.b64decode(b64str.encode('utf8'))
    data = np.fromstring(data, np.uint8)
    data = cv2.imdecode(data, cv2.IMREAD_COLOR)
    return data


def draw_boxes(image, boxes, scores=None, drop_score=0.5):
    if scores is None:
        scores = [1] * len(boxes)
    for (box, score) in zip(boxes, scores):
        if score < drop_score:
            continue
        box = np.reshape(np.array(box), [-1, 1, 2]).astype(np.int64)
        image = cv2.polylines(np.array(image), [box], True, (255, 0, 0), 2)
    return image


def get_rotate_crop_image(img, points):
    '''
    img_height, img_width = img.shape[0:2]
    left = int(np.min(points[:, 0]))
    right = int(np.max(points[:, 0]))
    top = int(np.min(points[:, 1]))
    bottom = int(np.max(points[:, 1]))
    img_crop = img[top:bottom, left:right, :].copy()
    points[:, 0] = points[:, 0] - left
    points[:, 1] = points[:, 1] - top
    '''
    assert len(points) == 4, "shape of points must be 4*2"
    img_crop_width = int(
        max(
            np.linalg.norm(points[0] - points[1]),
            np.linalg.norm(points[2] - points[3])))
    img_crop_height = int(
        max(
            np.linalg.norm(points[0] - points[3]),
            np.linalg.norm(points[1] - points[2])))
    pts_std = np.float32([[0, 0], [img_crop_width, 0],
                          [img_crop_width, img_crop_height],
                          [0, img_crop_height]])
    M = cv2.getPerspectiveTransform(points, pts_std)
    dst_img = cv2.warpPerspective(
        img,
        M, (img_crop_width, img_crop_height),
        borderMode=cv2.BORDER_REPLICATE,
        flags=cv2.INTER_CUBIC)
    dst_img_height, dst_img_width = dst_img.shape[0:2]
    if dst_img_height * 1.0 / dst_img_width >= 1.5:
        dst_img = np.rot90(dst_img)
    return dst_img


def check_gpu(use_gpu):
    if use_gpu and not paddle.is_compiled_with_cuda():

        use_gpu = False
    return use_gpu


if __name__ == '__main__':
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/11/05 09:01

"""
pp是paddlepaddle的缩写
"""

import os
import sys
import logging
import random
import shutil
import re

from tqdm import tqdm
import numpy as np
import pandas as pd
import humanfriendly

import paddle
import paddle.inference as paddle_infer
import paddle.io
# from paddle.io import DataLoader, Dataset

from pyxllib.algo.pupil import natural_sort
from pyxllib.xl import XlPath, browser
from pyxllib.xlcv import xlcv
from pyxlpr.ai.specialist import ClasEvaluater, show_feature_map


def __1_数据集():
    pass


class SequenceDataset(paddle.io.Dataset):
    def __init__(self, samples, labels=None, transform=None):
        super().__init__()
        self.samples = samples
        self.labels = labels
        # if self.labels is not None:  # 对np.array类型无法直接使用len，从通用角度看，这个参数可以不设
        #     assert len(self.samples) == len(self.labels)
        self.transform = transform

    def __len__(self):
        return len(self.samples)

    def __getitem__(self, index):
        x = self.samples[index]
        if self.transform:
            x = self.transform(x)

        if self.labels is not None:
            return x, self.labels[index]
        else:
            return x 


def build_testdata_loader(samples, *, labels=None, transform=None, **kwargs):
    """
    简化的一个创建paddle的DataLoader的函数。主要用于简化部署阶段的推理。
    """
    if isinstance(samples, paddle.io.DataLoader):
        return samples
    elif isinstance(samples, paddle.io.Dataset):
        dataset = samples
    else:
        dataset = SequenceDataset(samples, labels, transform)

    return paddle.io.DataLoader(dataset, **kwargs)


class ImageClasDataset(paddle.io.Dataset):
    """ 常用的分类数据集格式 """

    def __init__(self, num_classes, samples, ratio=None, *,
                 use_img_augment=False, seed=4101, class_names=None):
        """ 直接按root下的目录数进行分类，注意如果train和val是分开的，目录结构要一直，否则自带的类别id编号会不匹配

        :param num_classes: 类别数
        :param list samples: 样本清单，每个条目有[图片路径, 类别id]
        :param int|float|list|tuple ratio: 取数据的比例，默认全取，可以输入一个区间，指定取哪个部分
            这个操作会设置seed，确保每次随机打乱后选取的结果相同
        :param list class_names: 表示id从0开始依次取，对应的类别昵称
        """
        super().__init__()

        if ratio is not None:
            if isinstance(ratio, (int, float)):
                # 每个类别取得样本区间
                if ratio > 0:
                    left, right = 0, ratio
                else:
                    left, right = (1 + ratio), 1
            else:
                left, right = ratio

            # 初始化，按类别分好组
            random.seed(seed)
            groups = [[] for i in range(num_classes)]
            for file, label in samples:
                groups[label].append(file)

            # 每个类别选取部分数据
            samples = []
            for label, files in enumerate(groups):
                n = len(files)
                random.shuffle(files)
                files2 = files[int(left * n):int(right * n)]
                samples += [[f, label] for f in files2]

        self.samples = samples
        self.num_classes = num_classes
        self.class_names = class_names
        self.use_img_augment = use_img_augment

    @classmethod
    def from_folder(cls, root, ratio=None, *, class_mode=1, **kwargs):
        """ 从类别目录式的数据，构造图像分类数据集

        :param root: 数据根目录
        :param ratio: 每个类别取多少样本量
        :param class_mode: 类别限定方法。注意空目录也会标记为1个类。
            0，一般是读取没有label标签的测试集，所有的类别，统一用0占位
            1，root下每个直接子目录是一个类别，每个类别目录里如果有嵌套目录，都会归为可用图片
            2，root下每个目录均被视为一个类别，这些类别在目录结构上虽然有嵌套结构，但在模型上先用线性类别模式处理

        注：空目录相当于没有该类别数据，会跳过，不会进入分类清单。比如8个类别的目录，但是有2个空的，那么实际只会生成6分类模型。
        """

        def run_mode0():
            samples = list(XlPath(root).glob_images('**/*'))
            return samples, []

        def run_mode1():
            samples, class_names = [], []
            dirs = sorted(XlPath(root).glob_dirs())
            for i, d in enumerate(dirs):
                class_names.append(d.name)
                for f in d.glob_images('**/*'):
                    samples.append([f, i])
            return samples, class_names

        def run_mode2():
            samples, class_names = [], []
            dirs = sorted(XlPath(root).rglob_dirs())
            for i, d in enumerate(dirs):
                class_names.append(d.name)
                for f in d.glob_images():
                    samples.append([f, i])
            return samples, class_names

        func = {0: run_mode0, 1: run_mode1, 2: run_mode2}[class_mode]
        samples, class_names = func()
        return cls(len(class_names), samples, ratio, class_names=class_names, **kwargs)

    @classmethod
    def from_label(cls, label_file, root=None, ratio=None, **kwargs):
        """ 从标注文件初始化 """
        label_file = XlPath(label_file)
        lines = label_file.read_text().splitlines()
        if root is None:
            root = label_file.parent
        else:
            root = XlPath(root)

        samples, class_names = [], set()
        for line in lines:
            if not line:
                continue
            path, label = line.split('\t')
            class_names.add(label)
            samples.append([root / path, int(label)])

        class_names = natural_sort(list(class_names))

        return cls(len(class_names), samples, ratio=ratio, class_names=class_names, **kwargs)

    def __len__(self):
        return len(self.samples)

    def save_class_names(self, outfile):
        """ 保存类别文件 """
        class_names = self.class_names
        if not class_names:
            class_names = list(map(str, range(self.num_classes)))
        outfile = XlPath(outfile)
        if not outfile.parent.is_dir():
            os.makedirs(outfile.parent)
        outfile.write_text('\n'.join(class_names))

    @classmethod
    def img_augment(cls, img):
        """ 自带的一套默认的增广、数据处理方案。实际应用建议根据不同任务做扩展调整。
        """
        import albumentations as A
        h, w, c = img.shape
        # 如果进行随机裁剪，则h, w的尺寸变化
        h = random.randint(int(h * 0.7), h)
        w = random.randint(int(w * 0.7), w)
        transform = A.Compose([
            A.RandomCrop(width=w, height=h, p=0.8),
            A.CoarseDropout(),  # 随机噪声遮挡
            A.RandomSunFlare(p=0.1),  # 随机强光
            A.RandomShadow(p=0.1),  # 随机阴影
            A.RGBShift(p=0.1),  # RGB波动
            A.Blur(p=0.1),  # 模糊
            A.RandomBrightnessContrast(p=0.2),  # 随机调整图片明暗
        ])
        return transform(image=img)['image']

    @classmethod
    def transform(cls, x):
        """ 自带的一种默认的图片预处理方案，实际应用建议根据不同任务做扩展调整。
        """
        import paddle.vision.transforms.functional as F
        img = xlcv.read(x)
        img = F.resize(img, (256, 256))  # 将图片尺寸统一，方便按batch训练。但resnet并不强制输入图片尺寸大小。
        img = np.array(img, dtype='float32') / 255.
        img = img.transpose([2, 0, 1])
        return img

    def __getitem__(self, index):
        file, label = self.samples[index]
        img = xlcv.read(file)
        if self.use_img_augment:
            img = self.img_augment(img)
        img = self.transform(img)
        return img, np.array(label, dtype='int64')


def __2_模型结构():
    pass


def check_network(x):
    """ 检查输入的模型x的相关信息 """
    msg = '总参数量：'
    msg += str(sum([p.size for p in x.parameters()]))
    msg += ' | ' + ', '.join([f'{p.name}={p.size}' for p in x.parameters()])
    print(msg)


def model_state_dict_df(model, *, browser=False):
    """ 统计模型中所有的参数

    :param browser: 不单纯返回统计表，而是用浏览器打开，展示更详细的分析报告

    详细见 w211206周报
    """
    ls = []
    # 摘选ParamBase中部分成员属性进行展示
    columns = ['var_name', 'name', 'shape', 'size', 'dtype', 'trainable', 'stop_gradient']

    state_dict = model.state_dict()  # 可能会有冗余重复

    used = set()
    for k, v in state_dict.items():
        # a 由于state_dict的机制，self.b=self.a，a、b都是会重复获取的，这时候不应该重复计算参数量
        # 但是后面计算存储文件大小的时候，遵循原始机制冗余存储计算空间消耗
        param_id = id(v)
        if param_id in used:
            continue
        else:
            used.add(param_id)
        # b msg
        msg = [k]
        for col_name in columns[1:]:
            msg.append(getattr(v, col_name, None))
        ls.append(msg)
    df = pd.DataFrame.from_records(ls, columns=columns)

    def html_content(df):
        import io

        content = f'<pre>{model}' + '</pre><br/>'
        content += df.to_html()
        total_params = sum(df['size'])
        content += f'<br/>总参数量：{total_params}'

        f = io.BytesIO()
        paddle.save(state_dict, f)
        content += f'<br/>文件大小：{humanfriendly.format_size(len(f.getvalue()))}'
        return content

    if browser:
        browser.html(html_content(df))

    return df


def __3_损失():
    pass


def __4_优化器():
    pass


def __5_评价指标():
    pass


class ClasAccuracy(paddle.metric.Metric):
    """ 分类问题的精度 """

    def __init__(self, num_classes=None, *, print_mode=0):
        """
        :param num_classes: 其实这个参数不输也没事~~
        :param print_mode:
            0，静默
            1，reset的时候，输出f1指标
            2，reset的时候，还会输出crosstab
        """
        super(ClasAccuracy, self).__init__()
        self.num_classes = num_classes
        self.total = 0
        self.count = 0
        self.gt = []
        self.pred = []
        self.print_mode = print_mode

    def name(self):
        return 'acc'

    def update(self, x, y):
        x = x.argmax(axis=1)
        y = y.reshape(-1)
        cmp = (x == y)
        self.count += cmp.sum()
        self.total += len(cmp)
        self.gt += y.tolist()
        self.pred += x.tolist()

    def accumulate(self):
        return self.count / self.total

    def reset(self):
        if self.print_mode:
            a = ClasEvaluater(self.gt, self.pred)
            print(a.f1_score('all'))
            if self.print_mode > 1:
                print(a.crosstab())
        self.count = 0
        self.total = 0
        self.gt = []
        self.pred = []


class VisualAcc(paddle.callbacks.Callback):
    def __init__(self, logdir, experimental_name, *, reset=False, save_model_with_input=None):
        """
        :param logdir: log所在根目录
        :param experimental_name: 实验名子目录
        :param reset: 是否重置目录
        :param save_model_with_input: 默认不存储模型结构
        """
        from visualdl import LogWriter

        super().__init__()
        # 这样奇怪地加后缀，是为了字典序后，每个实验的train显示在eval之前
        d = XlPath(logdir) / (experimental_name + '_train')
        if reset and d.exists(): shutil.rmtree(d)
        self.write = LogWriter(logdir=str(d))
        d = XlPath(logdir) / (experimental_name + '_val')
        if reset and d.exists(): shutil.rmtree(d)
        self.eval_writer = LogWriter(logdir=str(d))
        self.eval_times = 0

        self.save_model_with_input = save_model_with_input

    def on_epoch_end(self, epoch, logs=None):
        self.write.add_scalar('acc', step=epoch, value=logs['acc'])
        self.write.flush()

    def on_eval_end(self, logs=None):
        self.eval_writer.add_scalar('acc', step=self.eval_times, value=logs['acc'])
        self.eval_writer.flush()
        self.eval_times += 1


def __6_集成():
    pass


class XlModel(paddle.Model):
    def __init__(self, network, **kwargs):
        """

        """
        super(XlModel, self).__init__(network, **kwargs)
        self.save_dir = None
        self.train_data = None
        self.eval_data = None
        self.test_data = None
        self.callbacks = []

    def get_save_dir(self):
        """
        注意 self.save_dir、self.get_save_dir()各有用途
            self.save_dir获取原始配置，可能是None，表示未设置，则在某些场合默认不输出文件
            self.get_save_dir()，有些场合显示指定要输出文件了，则需要用这个接口获得一个明确的目录
        """
        if self.save_dir is None:
            return XlPath('.')
        else:
            return self.save_dir

    def set_save_dir(self, save_dir):
        """
        :param save_dir: 模型等保存的目录，有时候并不想保存模型，则可以不设
            如果在未设置save_dir情况下，仍使用相关读写文件功能，默认在当前目录下处理
        """
        # 相关数据的保存路径
        self.save_dir = XlPath(save_dir)
        os.makedirs(self.save_dir, exist_ok=True)

    def set_dataset(self, train_data=None, eval_data=None, test_data=None):
        if train_data:
            self.train_data = train_data
            if isinstance(train_data, ImageClasDataset):
                self.train_data.save_class_names(self.get_save_dir() / 'class_names.txt')  # 保存类别昵称文件
        if eval_data:
            self.eval_data = eval_data
        if test_data:
            self.test_data = test_data
            # TODO 可以扩展一些能自动处理测试集的功能
            #   不过考虑不同性质的任务，这个不太好封装，可能要分图像分类，目标检测的分类处理
            #   但这样一来就等于重做一遍PaddleDet等框架了，不太必要

    def try_load_params(self, relpath='final.pdparams'):
        # TODO 怎么更好地resume训练？回复学习率等信息？虽然目前直接加载权重重新训练也没大碍。
        pretrained_model = self.get_save_dir() / relpath
        if pretrained_model.is_file():
            self.network.load_dict(paddle.load(str(pretrained_model)))

    def prepare_clas_task(self, optimizer=None, loss=None, metrics=None, amp_configs=None,
                          use_visualdl=None):
        """ 分类模型的一套默认的优化器、损失、测评配置

        TODO 这套配置不一定是最泛用的，需要进行更多研究

        :param use_visualdl: 是否使用visualdl
            支持输入str类型，作为自定义路径名
            否则每次实验，会自增一个编号，生成 e0001、e0002、e0003、...
        """
        from paddle.optimizer import Momentum
        from paddle.regularizer import L2Decay

        if optimizer is None:
            optimizer = Momentum(learning_rate=0.01,
                                 momentum=0.9,
                                 weight_decay=L2Decay(1e-4),
                                 parameters=self.network.parameters())
        if loss is None:
            loss = paddle.nn.CrossEntropyLoss()
        if metrics is None:
            metrics = ClasAccuracy(print_mode=2)  # 自定义可以用crosstab检查的精度类

        self.prepare(optimizer, loss, metrics, amp_configs)

        # 但有设置save_dir的时候，默认开启可视化
        if use_visualdl is None and self.save_dir is not None:
            use_visualdl = True

        if use_visualdl:
            p = self.save_dir or XlPath('.')
            if not isinstance(use_visualdl, str):
                num = max([int(re.search(r'\d+', x.stem).group())
                           for x in p.glob_dirs()
                           if re.match(r'e\d+_', x.stem)], default=0) + 1
                use_visualdl = f'e{num:04}'
            self.callbacks.append(VisualAcc(p / 'visualdl', use_visualdl))

    def train(self,
              epochs=1,
              batch_size=1,
              eval_freq=1000,  # 每多少次epoch进行精度验证，可以调大些，默认就是不验证了。反正目前机制也没有根据metric保存最优模型的操作。
              log_freq=1000,  # 每轮epoch中，每多少step显示一次日志，可以调大些
              save_freq=1000,  # 每多少次epoch保存模型。可以调大些，默认就只保存final了。
              verbose=2,
              drop_last=False,
              shuffle=True,
              num_workers=0,
              callbacks=None,
              accumulate_grad_batches=1,
              num_iters=None,
              ):
        """ 对 paddle.Model.fit的封装

        简化了上下游配置
        修改了一些参数默认值，以更符合我实际使用中的情况
        """
        train_data = self.train_data
        eval_data = self.eval_data

        callbacks = callbacks or []
        if self.callbacks:
            callbacks += self.callbacks

        super(XlModel, self).fit(train_data, eval_data, batch_size, epochs, eval_freq, log_freq,
                                 self.save_dir, save_freq, verbose, drop_last, shuffle, num_workers,
                                 callbacks, accumulate_grad_batches, num_iters)

        # 判断最后是否要再做一次eval：有验证集 + 原本不是每次epoch都预测 + 正好最后次epochs结束是eval周期结束
        # 此时paddle.Model.fit机制是恰好不会做eval的，这里做个补充
        if eval_data and eval_freq != 1 and (epochs % eval_freq == 0):
            self.evaluate(eval_data)

        # TODO 要再写个metric测评？这个其实就是evaluate，不用重复写吧。

    def save_static_network(self, *, data_shape=None):
        """ 导出静态图部署模型 """
        if data_shape is None:
            data_shape = [1, 3, 256, 256]
            # TODO 可以尝试从train_data、eval_data等获取尺寸
        data = paddle.zeros(data_shape, dtype='float32')
        infer_dir = self.get_save_dir() / 'infer/inference'
        infer_dir.mkdir(parents=True, exist_ok=True)
        paddle.jit.save(paddle.jit.to_static(self.network), infer_dir.as_posix(), [data])


def __7_部署():
    pass


class ImageClasPredictor:
    """ 图像分类框架的预测器 """

    def __init__(self, model, *, transform=None, class_names=None):
        self.model = model
        self.transform = transform
        # 如果输入该字段，会把下标id自动转为明文类名
        self.class_names = class_names

    @classmethod
    def from_dynamic(cls, model, params_file=None, **kwargs):
        """ 从动态图初始化 """
        if params_file:
            model.load_dict(paddle.load(params_file))
        model.eval()
        return cls(model, **kwargs)

    @classmethod
    def from_static(cls, pdmodel, pdiparams, **kwargs):
        """ 从静态图初始化 """
        # 创建配置对象，并根据需求配置
        config = paddle_infer.Config(pdmodel, pdiparams)
        device = paddle.get_device()

        if device.startswith('gpu'):
            config.enable_use_gpu(0, int(device.split(':')[1]))

        # 根据Config创建预测对象
        predictor = paddle_infer.create_predictor(config)

        def model(x):
            """ 静态图的使用流程会略麻烦一点

            以及为了跟动态图的上下游衔接，需要统一格式
                输入的tensor x 需要改成 np.array
                输出的np.array 需要改成 tensor

            TODO 关于这里动静态图部署的代码，可能有更好的组织形式，这个以后继续研究吧~~
            """
            # 获取输入的名称
            input_names = predictor.get_input_names()
            # 获取输入handle
            x_handle = predictor.get_input_handle(input_names[0])
            x_handle.copy_from_cpu(x.numpy())
            # 运行预测引擎
            predictor.run()
            # 获得输出名称
            output_names = predictor.get_output_names()
            # 获得输出handle
            output_handle = predictor.get_output_handle(output_names[0])
            output_data = output_handle.copy_to_cpu()  # return numpy.ndarray
            return paddle.Tensor(output_data)

        return cls(model, **kwargs)

    @classmethod
    def from_modeldir(cls, root, *, dynamic_net=None, **kwargs):
        """ 从特定的目录结构中初始化部署模型
        使用固定的配置范式，我自己常用的训练目录结构

        :param dynamic_net: 输入动态图模型类型，初始化动态图

        注：使用这个接口初始化，在目录里必须要有个class_names.txt文件来确定类别数
            否则请用更底层的from_dynamic、from_static精细配置
        """
        root = XlPath(root)
        class_names_file = root / 'class_names.txt'
        assert class_names_file.is_file(), f'{class_names_file} 必须要有类别昵称配置文件，才知道类别数'
        class_names = class_names_file.read_text().splitlines()

        if dynamic_net:
            clas = ImageClasPredictor.from_dynamic(dynamic_net(num_classes=len(class_names)),
                                                   str(root / 'final.pdparams'),
                                                   class_names=class_names,
                                                   **kwargs)
        else:
            clas = cls.from_static(str(root / 'infer/inference.pdmodel'),
                                   str(root / 'infer/inference.pdiparams'),
                                   class_names=class_names,
                                   **kwargs)

        return clas

    def pred_batch(self, samples, batch_size=None, *, return_mode=0, print_mode=0):
        """ 默认是进行批量识别，如果只识别单个，可以用pred

        :param samples: 要识别的数据，支持类list的列表，或Dataset、DataLoader
        :param return_mode: 返回值细粒度，0表示直接预测类别，1则是返回每个预测在各个类别的概率
        :param print_mode: 0 静默运行，1 显示进度条
        :param batch_size: 默认按把imgs整个作为一个批次前传，如果数据量很大，可以使用该参数切分batch
        :return:
        """
        import paddle.nn.functional as F

        if not batch_size: batch_size = len(samples)
        data_loader = build_testdata_loader(samples, transform=self.transform, batch_size=batch_size)

        logits = []
        for inputs in tqdm(data_loader, desc='预测：', disable=not print_mode):
            logits.append(self.model(inputs))
            # if sys.version_info.minor >= 8:  # v0.1.62.2 paddlelib bug，w211202
            #     break
        logits = paddle.concat(logits, axis=0)

        if return_mode == 0:
            idx = logits.argmax(1).tolist()
            if self.class_names:
                idx = [self.class_names[x] for x in idx]
            return idx
        elif return_mode == 1:
            prob = F.softmax(logits, axis=1).tolist()
            for i, item in enumerate(prob):
                prob[i] = [round(x, 4) for x in item]  # 保留4位小数就够了
            return prob
        else:
            raise ValueError

    def __call__(self, *args, **kwargs):
        return self.pred_batch(*args, **kwargs)

    def pred(self, img, *args, **kwargs):
        return self.pred_batch([img], *args, **kwargs)[0]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2025/06/05

""" 这是一套基于labelme标注来进行 """

import sys
from collections import defaultdict
import json
import os
import time
import random
import re
import datetime
from collections import UserDict

import numpy as np
import win32com
from loguru import logger
import pyscreeze

from pyxlpr.ai.clientlib import XlAiClient

if sys.platform == 'win32':
    import pyautogui
    import win32gui

from pyxllib.prog.newbie import first_nonnone, round_int
from pyxllib.prog.pupil import xlwait, DictTool, run_once
from pyxllib.prog.filelock import get_autogui_lock
from pyxllib.algo.geo import ComputeIou, ltrb2xywh, xywh2ltrb
from pyxllib.file.specialist import XlPath
from pyxllib.cv.expert import xlcv, xlpil
from pyxlpr.data.labelme import LabelmeDict
from pyxllib.autogui.uiautolib import find_ctrl, UiCtrlNode

"""
桌面自动化操作，Anto GUI，简称An

相关概念，名称定义如下

从空间位置理解：
monitor：实际存在的若干物理屏幕
screen：多个monitor组合在一起的整个逻辑屏幕
region：screen中的某个局部区域
window: 特指某个软件所处的区域

从状态(时间)理解：
view: 同一个窗口位置上，可能有不同的各种功能、菜单变化，称为不同的视图
shape: view里的各个位置对象标记内容。上述所有类其实都是shape类
"""


@run_once()
def get_xlapi():
    xlapi = XlAiClient(auto_login=False, check=False)
    logger.info(f"{os.getenv('XL_API_PRIU_TOKEN')}, {os.getenv('MAIN_WEBSITE')}")
    xlapi.login_priu(os.getenv('XL_API_PRIU_TOKEN'), os.getenv('MAIN_WEBSITE'))
    return xlapi


class ImageTools:

    @classmethod
    def pixel_distance(cls, pixel1, pixel2):
        """ 返回最大的像素值差，如果相同返回0 """
        return max([abs(x - y) for x, y in zip(pixel1, pixel2)])

    @classmethod
    def img_distance(cls, img1, img2, *, grayscale=None, color_tolerance=10):
        """ 返回距离：不相同像素占的百分比，相同返回0

        :param grayscale: 是否转换为灰度图进行比较，默认为None，即不转换
        :param color_tolerance: 每个像素允许的颜色差值。RGB情况计算的是3个值分别的差的和。
        """
        img1 = np.array(img1, dtype=int)
        img2 = np.array(img2, dtype=int)

        if grayscale:
            # 如果是灰度图，将图像转换为单通道
            if len(img1.shape) == 3:
                img1 = np.mean(img1, axis=2, dtype=int)
            if len(img2.shape) == 3:
                img2 = np.mean(img2, axis=2, dtype=int)
            cmp = np.array(abs(img1 - img2) > color_tolerance)
        else:
            # 彩色图情况，计算每个通道的差值
            cmp = np.array(abs(img1 - img2) > color_tolerance)
            # 只要有一个通道的差值超过阈值，就认为该像素不同
            cmp = np.any(cmp, axis=-1)

        return cmp.sum() / cmp.size

    @classmethod
    def base_find_img(cls, img, haystack=None, *, grayscale=None, confidence=None, sort_by_confidence=False):
        """ 根据预存的img数据，匹配出多个内容对应的所在的rect位置
        这函数不能取同名find_img，否则IDE不好自动识别跳转

        :param img: 要查找的目标的子图
        :param haystack: 整张大图
        :param grayscale: 是否使用灰度图像进行匹配
        :param confidence: 匹配置信度阈值
        :param sort_by_confidence: 是否按照置信度排序返回结果
        :return: 会返回多个匹配结果的坐标
        """
        # 若不需要按置信度排序，使用原始pyautogui方法
        if not sort_by_confidence:
            return cls._find_with_pyautogui(img, haystack, grayscale, confidence)
        else:
            return cls._find_with_opencv(img, haystack, grayscale, confidence)

    @classmethod
    def _find_with_pyautogui(cls, img, haystack, grayscale, confidence):
        """使用pyautogui原始方法查找匹配"""
        try:
            boxes = pyautogui.locateAll(img, haystack, grayscale=grayscale, confidence=confidence)

            try:
                boxes = list(boxes)
            except pyscreeze.ImageNotFoundException:
                return []

            # 过滤掉重叠超过一半面积的框
            rects = ComputeIou.nms_xywh(boxes)

            # rects里会有numpy.int64类型的数据，需要做个转换
            rects2 = []
            for rect in rects:
                rects2.append([int(x) for x in rect])
            return rects2
        except pyautogui.ImageNotFoundException:
            # 捕获图像未找到异常，返回空列表
            return []

    @classmethod
    def _find_with_opencv(cls, img, haystack, grayscale, confidence):
        """使用OpenCV查找匹配并按置信度排序"""
        try:
            # 转换图像格式
            template, search_img = cls._prepare_images(img, haystack, grayscale)

            # 获取匹配结果及置信度
            matches_with_conf = cls._match_template_with_confidence(template, search_img, confidence)

            # 按置信度排序
            matches_with_conf.sort(key=lambda x: x[4], reverse=True)

            # 提取位置信息
            boxes = [(x, y, w, h) for x, y, w, h, _ in matches_with_conf]

            # 应用NMS
            if boxes:
                rects = ComputeIou.nms_xywh(boxes)

                # 转换为整数类型
                rects2 = []
                for rect in rects:
                    rects2.append([int(x) for x in rect])
                return rects2
            return []
        except Exception as e:
            # 捕获异常，返回空列表
            print(f"Error in OpenCV template matching: {e}")
            return []

    @classmethod
    def _prepare_images(cls, img, haystack, grayscale):
        """准备图像用于OpenCV模板匹配"""
        import cv2
        import numpy as np

        # 如果是文件路径，则读取图像
        if isinstance(img, str):
            template = cv2.imread(img, cv2.IMREAD_COLOR)
        elif isinstance(img, np.ndarray):
            template = img
        else:
            # 如果是PIL图像，转换为OpenCV格式
            template = np.array(img)
            template = cv2.cvtColor(template, cv2.COLOR_RGB2BGR)

        if isinstance(haystack, str):
            search_img = cv2.imread(haystack, cv2.IMREAD_COLOR)
        elif isinstance(haystack, np.ndarray):
            search_img = haystack
        else:
            # 如果是PIL图像，转换为OpenCV格式
            search_img = np.array(haystack)
            search_img = cv2.cvtColor(search_img, cv2.COLOR_RGB2BGR)

        # 如果需要灰度图像
        if grayscale:
            if len(template.shape) > 2:
                template = cv2.cvtColor(template, cv2.COLOR_BGR2GRAY)
            if len(search_img.shape) > 2:
                search_img = cv2.cvtColor(search_img, cv2.COLOR_BGR2GRAY)

        return template, search_img

    @classmethod
    def _match_template_with_confidence(cls, template, search_img, confidence):
        """使用OpenCV进行模板匹配并返回带置信度的结果"""
        import cv2
        import numpy as np

        # 确保置信度设置
        if confidence is None:
            confidence = 0.8

        # 执行模板匹配
        result = cv2.matchTemplate(search_img, template, cv2.TM_CCOEFF_NORMED)
        w, h = template.shape[1], template.shape[0]

        # 找出所有超过阈值的点
        locations = np.where(result >= confidence)
        matches = []

        # 收集坐标和置信度
        for pt in zip(*locations[::-1]):  # 转换为(x, y)格式
            x, y = pt
            conf = float(result[y, x])  # 获取该位置的置信度值
            matches.append((x, y, w, h, conf))

        return matches


class _AnShapeBasic(UserDict):
    """ 基本你的初始化、保存等模块 """

    def __1_构建(self):
        pass

    def __init__(self, initialdata=None, parent=None):
        """
        :param initialdata: 这个字典基础key字段和含义如下
            text，原本label标注的纯文本信息
            xywh，相对parent所处的外接矩形坐标位置
            center，中心点坐标

            （使用图片初始化的时候才有以下字段）
            img，cv格式的图片数据（这是特指labelme中存储的静态数据）
            pixel，中心点的像素值

            shot，这是特指程序运行中实时截图获取的动态快照图片
        :param parent:
        """
        super().__init__(initialdata)
        self.parent = parent or None

    def set_shape(self, text, xywh):
        """ 配置形状 """
        self['text'] = text
        self['xywh'] = xywh
        # 通过xywh计算出center
        self['center'] = [xywh[0] + xywh[2] // 2, xywh[1] + xywh[3] // 2]

    def read_lmshape(self, lmshape):
        """
        将labelme格式的shape字典改为该任务特有的字典结构
        """
        # 1 解析原label为字典
        # 如果是来自普通的label，会把原文本自动转换为text字段。如果是来自xllabelme，则一般默认就是字典值，带text字段。
        if isinstance(lmshape['label'], str):
            anshape = DictTool.json_loads(lmshape['label'], 'text')
        else:
            anshape = lmshape['label']
        anshape.update(DictTool.sub(lmshape, ['label']))  # 因为原本的label被展开成字典了，这里要删掉label

        # 2 中心点 center（无论任何原始形状，都转成矩形处理。这跟下游功能有关，目前只能做矩形。）
        shape_type = lmshape['shape_type']
        pts = lmshape['points']
        if shape_type in ('rectangle', 'polygon', 'line'):
            anshape['center'] = np.array(np.array(pts).mean(axis=0), dtype=int).tolist()
        elif shape_type == 'circle':
            anshape['center'] = pts[0]

        # 3 外接矩形 rect
        if shape_type in ('rectangle', 'polygon'):
            ltrb = np.array(pts, dtype=int).reshape(-1).tolist()
        elif shape_type == 'circle':
            x, y = pts[0]
            r = ((x - pts[1][0]) ** 2 + (y - pts[1][1]) ** 2) ** 0.5
            ltrb = [round_int(v) for v in [x - r, y - r, x + r, y + r]]
        # 把ltrb转换为xywh
        anshape['xywh'] = ltrb2xywh(ltrb)

        # 4 图片数据 img, etag
        if self.parent.get('img') is not None and anshape['xywh']:
            anshape['img'] = xlcv.get_sub(self.parent['img'], ltrb)
            # anshape['etag'] = get_etag(anshape['img'])
            # TODO 这里目的其实就是让相同图片对比尽量相似，所以可以用dhash而不是etag

        # 5 中心点像素值 pixel
        p = anshape['center']
        if self.parent.get('img') is not None and p:
            anshape['pixel'] = tuple(self.parent['img'][p[1], p[0]].tolist()[::-1])

        self.update(anshape)

    def __2_转换与保存(self):
        pass

    def convert_to_view(self, view_name=None):
        """ 将这个shape升级为一个view对象 """
        img = self.shot()
        view = AnView.init_from_image(img)
        view.parent = self
        view['text'] = view_name
        return view

    @classmethod
    def raw_save_image(cls, img, region_folder, view_name=None, timetag=None):
        timetag_ = datetime.datetime.now().strftime('%Y%m%d_%H%M%S_%f')[:-3]
        if not view_name:  # 未输入则用时间戳代替
            view_name = timetag_
        elif timetag:  # 如果指定了时间戳参数，则强制加上时间戳前缀
            view_name = timetag_ + view_name

        impath = XlPath(region_folder) / f'{view_name}.jpg'
        os.makedirs(os.path.dirname(impath), exist_ok=True)
        xlcv.write(img, impath)

    def save_image(self, region_folder, view_name=None, timetag=None):
        """ 保存当前快照图片 """
        self.raw_save_image(self.shot(), region_folder, view_name, timetag)

    def save_view(self, region_folder, view_name=None, timetag=None):
        """ 保存当前视图（即保存图片和对应的标注数据） """
        view = self.convert_to_view(view_name)

        timetag_ = datetime.datetime.now().strftime('%Y%m%d_%H%M%S_%f')[:-3]
        if not view_name:  # 未输入则用时间戳代替
            view_name = timetag_
        elif timetag:  # 如果指定了时间戳参数，则强制加上时间戳前缀
            view_name = timetag_ + view_name

        impath = XlPath(region_folder) / f'{view_name}.jpg'
        view.save_labelme_file(impath)

    def __3_基础功能(self):
        pass

    def get_parent_argv(self, arg_name, cur_value=None):
        """ 查找arg_name的配置值
        从当前节点往父节点找，找到第1个非None值作为配置
        """
        cur_shape = self
        while cur_value is None:
            cur_value = cur_shape.get(arg_name)
            if cur_value is not None:
                break
            elif cur_shape.parent is None:
                break
            else:
                cur_shape = cur_shape.parent
        return cur_value

    def get_abs_point(self, point):
        """
        将相对坐标转换为绝对坐标

        :param point: 相对于当前节点的坐标，默认为None（使用自身中心点）
        :return: 绝对坐标 [x, y]
        """
        point = point.copy()

        # 向上遍历所有父节点，累加偏移量
        current = self
        while current.parent and current.parent.get('xywh'):
            # 父节点的xywh格式为[x, y, width, height]
            x, y = current.parent['xywh'][:2]
            # 将父节点的偏移量添加到绝对坐标
            point[0] += x
            point[1] += y
            # 移动到上一级父节点
            current = current.parent

        return point

    def total_name(self):
        """ 按父节点 text1/text2 的形式组织的完整text内容 """
        names = []
        cur = self
        while cur:
            names.append(cur.get('text', ''))
            cur = cur.parent
        return '/'.join(names[::-1])

    def __3_获取图像(self):
        pass

    def shot(self):
        """ 截取当前的快照画面 """
        # 计算绝对坐标下的xywh
        x, y = self.get_abs_point(self['xywh'][:2])  # 左上角相对坐标转绝对坐标
        w, h = self['xywh'][2:]  # 宽度和高度
        # logger.info(f'{x} {y} {w} {h}')
        pil_img = pyautogui.screenshot(region=[x, y, w, h])
        return xlpil.to_cv2_image(pil_img)

    def get_pixel(self):
        """ 获得当前中心点的像素值

         官方pyautogui.pixel版本，在windows有时候会出bug
         OSError: windll.user32.ReleaseDC failed : return 0
        """
        w, h = self['xywh'][2:]
        shot = self.shot()
        return tuple(shot[h // 2, w // 2].tolist()[::-1])


class _AnShapePupil(_AnShapeBasic):
    """ 基础操作功能 """

    def __1_ocr识别(self):
        pass

    def ocr_text(self):
        text = get_xlapi().rec_singleline(self.shot())
        return text

    def ocr_value(self):
        vals = self.ocr_values()
        return vals[0] if vals else 0

    def ocr_values(self):
        def parse_val(v):
            return float(v) if '.' in v else int(v)

        text = self.ocr_text()
        vals = re.findall(r'\d+', text) or []
        vals = [parse_val(v) for v in vals]
        return vals

    def __2_基础操作(self):
        pass

    def move_to(self, *, random_bias=None):
        """ 移动鼠标到该形状的中心位置，支持随机偏移

        :param int random_bias: 允许在一定范围内随机偏移目标坐标
        """
        # 0 检索配置参数
        random_bias = self.get_parent_argv('random_bias', random_bias) or 0

        # 1 计算出相对根节点的目标point位置
        point = self.get_abs_point(self['center'])

        # 2 添加随机偏移
        if random_bias:
            point[0] += random.randint(-random_bias, random_bias)
            point[1] += random.randint(-random_bias, random_bias)

        # 3 执行鼠标移动
        with get_autogui_lock():  # 全局ui锁，避免跟微信同时操作等冲突问题
            pyautogui.moveTo(*point)
        return point

    def click(self, x_bias=0, y_bias=0, *, random_bias=None, back=None, wait_change=None, wait_seconds=None):
        """ 点击这个shape

        :param x_bias: 原本中心点击位置，增设一个偏移量
        :param y_bias: y轴偏移量
        :param int random_bias: 允许在一定范围内随机点击点
        :param bool back: 是否点击后鼠标移回原坐标
        :param wait_change: 点击后，等待画面产生变化，才结束函数
            注意这个只是比较局部区域图片变化，不是全图的变化
        :param wait_seconds: 点击后要继续等待若干秒再退出函数
        """
        # 0 检索配置参数
        random_bias = self.get_parent_argv('random_bias', random_bias) or 0
        back = self.get_parent_argv('back', back) or False
        wait_change = self.get_parent_argv('wait_change', wait_change) or False
        wait_seconds = self.get_parent_argv('wait_seconds', wait_seconds) or 0

        # 1 计算出相对根节点的目标point位置
        point = self.get_abs_point(self['center'])
        if x_bias or y_bias:
            point[0] += x_bias
            point[1] += y_bias

        # 2 添加随机偏移
        if random_bias:
            point[0] += random.randint(-random_bias, random_bias)
            point[1] += random.randint(-random_bias, random_bias)

        # 3 鼠标移动
        with get_autogui_lock():
            origin_point = pyautogui.position()
            if wait_change:
                before_click_shot = self.shot()
            pyautogui.click(*point)
            if back:
                pyautogui.moveTo(*origin_point)  # 恢复鼠标原位置

        # 4 等待
        if wait_change:
            color_tolerance = self.get_parent_argv('color_tolerance', None) or 10
            confidence = self.get_parent_argv('confidence', None) or 0.95
            func = lambda: ImageTools.img_distance(before_click_shot, self.shot(),
                                                   color_tolerance=color_tolerance) > confidence
            xlwait(func)
        if wait_seconds:
            time.sleep(wait_seconds)

        return point

    def drag_to(self, percent=50, direction='up', duration=None, **kwargs):
        """ 在当前shape里拖拽

        :param int percent: 拖拽比例，范围0-100，表示滚动区域的百分比
        :param str direction: 滚动方向，可选 'up', 'down', 'left', 'right'
        :param float duration: 拖拽动作持续时间(秒)
        """
        # 0 检索配置参数
        drag_duration = self.get_parent_argv('drag_duration', duration) or 1

        # 1 计算起点和终点位置
        x, y = self.get_abs_point(self['center'])
        width, height = self['xywh'][2:]

        # 限制百分比在0-100之间
        percent = max(0, min(100, percent))
        offset = percent / 2  # 从中心点偏移的百分比

        # 根据方向和百分比计算起点和终点
        if direction == 'up':
            # 从下往上拖
            start_point = [x, y + int(height * offset / 100)]
            end_point = [x, y - int(height * offset / 100)]
        elif direction == 'down':
            # 从上往下拖
            start_point = [x, y - int(height * offset / 100)]
            end_point = [x, y + int(height * offset / 100)]
        elif direction == 'left':
            # 从右往左拖
            start_point = [x + int(width * offset / 100), y]
            end_point = [x - int(width * offset / 100), y]
        elif direction == 'right':
            # 从左往右拖
            start_point = [x - int(width * offset / 100), y]
            end_point = [x + int(width * offset / 100), y]
        else:
            raise ValueError("方向必须是 'up', 'down', 'left', 'right' 之一")

        # 2 执行拖拽操作
        with get_autogui_lock():
            pyautogui.moveTo(*start_point)
            pyautogui.dragTo(*end_point, drag_duration, **kwargs)

        # 本来有想过返回拖拽前后图片是否有变化，但是这个会较影响性能，还是另外设计更合理


class AnShape(_AnShapePupil):
    """ 高级交互功能 """

    def __1_图像类(self):
        pass

    def find_img(self,
                 dst=None,
                 *,
                 part=None,
                 drag=0,
                 drag_percent=50,
                 drag_direction='up',
                 drag_duration=None,
                 color_tolerance=None,
                 grayscale=None,
                 confidence=None,
                 sort_by_confidence=False,
                 ):
        """ 查找图像匹配

        :param dst: 匹配目标
            None，默认自匹配，使用self['img']
        :param part:
            False, 全匹配模式，比如图片就是指整张图匹配，文本指整张图的文本
            True, 局部匹配模式，比如图片是指局部匹配到图片，文本指局部文本行匹配
            None, 根据上下文智能判断类型。
        :param int drag: 是否支持在该区域滚动检索，以及支持的拖拽上限次数
        :param color_tolerance: 每个像素允许的差值
            注意以下3个图像配参数，不要在函数里设默认值，函数里优先级是高于父节点设置参数的
            默认值只能在代码里get_parent_argv后再or默认值
        :param grayscale: 是否转成灰度图对比，默认不转。
        :param confidence: 置信度，距离。默认95%即要求95%区域的相同性。
        :return:
            全匹配模式返回None或self
            局部匹配返回匹配的shapes列表
        """
        # 1 匹配目标
        default_text = ''
        if dst is None:
            dst, part = self['img'], False
            default_text = self['text']
        elif isinstance(dst, str):  # XlPath不是str类型，利用这个特性可以输入图片路径用xlcv读取
            """ 如果输入字符串，表示这是一个相对当前self的view路径名 """
            dst = self[str]

        if isinstance(dst, AnShape):
            default_text = dst['text']
            img = dst['img']
            img_wh = dst['xywh'][:2]
            if part is None:
                part = self['xywh'][2:] != dst['xywh'][:2]
        else:
            # 否则当作输入自定义图片了，用xlcv读取
            img = xlcv.read(dst)
            img_wh = list(img.shape[1::-1])

        if part is None:
            # 根据wh判断part
            part = self['xywh'][2:] != img_wh

        # 配置参数
        color_tolerance = self.get_parent_argv('color_tolerance', color_tolerance) or 10
        grayscale = self.get_parent_argv('grayscale', grayscale) or False
        confidence = self.get_parent_argv('confidence', confidence) or 0.95

        # 2 完全匹配场景（该场景不支持drag参数）
        if not part:
            # 这种情况不用pyautogui的接口，用我自带的函数够了
            conf = 1 - ImageTools.img_distance(self.shot(), img, color_tolerance=color_tolerance)
            logger.info(f'{self.total_name()} {round(conf, 4)}')
            return self if conf > confidence else None

        # 3 局部匹配场景
        def find_subimg():
            # 单帧检索
            rects = ImageTools.base_find_img(img, self.shot(), grayscale=grayscale,
                                             confidence=confidence, sort_by_confidence=sort_by_confidence)
            # 把rects转成AnShape类型，每个rect都是[x, y, w, h]的list，不过其有numpy格式，要转成int类型
            shapes = []
            for rect in rects:
                sp = AnShape(parent=self)
                sp.set_shape(default_text, [int(x) for x in rect])
                shapes.append(sp)
            return shapes

        # 有无drag都走这套逻辑
        k, shapes = 0, find_subimg()
        while not shapes and k < drag:
            self.drag_to(drag_percent, drag_direction, drag_duration)
            shapes = find_subimg()
            k += 1

        return shapes

    def wait_img(self, dst=None, *, limit=None, interval=None, **kwargs):
        """ 等到目标匹配图片出现 """
        interval = self.get_parent_argv('wait_interval', interval) or 1
        return xlwait(lambda: self.find_img(dst, **kwargs), limit=limit, interval=interval)

    def waitleave_img(self, dst=None, *, limit=None, interval=None, **kwargs):
        """ 等到目标匹配图片离开（不再出现） """
        interval = self.get_parent_argv('wait_interval', interval) or 1
        return xlwait(lambda: not self.find_img(dst, **kwargs), limit=limit, interval=interval)

    def __2_文本类(self):
        """ find_系列统一返回shapes匹配列表 """
        pass

    def find_text(self,
                  dst=None,
                  *,
                  part=True,
                  drag=0,
                  drag_percent=50,
                  drag_direction='up',
                  drag_duration=None,
                  ):
        """ 查找文本匹配

        :param dst: 匹配目标
            None，默认自匹配，使用self['img']
        :param part:
            False, 全匹配模式，比如图片就是指整张图匹配，文本指整张图的文本
            True, 局部匹配模式，比如图片是指局部匹配到图片，文本指局部文本行匹配
        :param int drag: 是否支持在该区域滚动检索，以及支持的拖拽上限次数
        :return:
            全匹配模式返回None或self
            局部匹配返回匹配的shapes列表
        """
        # 1 匹配目标
        if dst is None:
            pattern = self['text']
        elif isinstance(dst, AnShape):
            pattern = dst['text']
        else:
            pattern = dst

        # 2 全图匹配
        if not part:
            return self if re.search(pattern, self.ocr_text()) else None

        # 3 局部匹配场景
        def find_subtext():
            shapes = []
            sub_view = self.convert_to_view()
            for sp in sub_view.shapes:
                if re.search(pattern, sp['text']):
                    shapes.append(sp)
            return shapes

        # 有无drag都走这套逻辑
        k, shapes = 0, find_subtext()
        while not shapes and k < drag:
            self.drag_to(drag_percent, drag_direction, drag_duration)
            shapes = find_subtext()
            k += 1

        return shapes

    def wait_text(self, dst=None, *, limit=None, interval=None, **kwargs):
        """ 等到目标匹配文本出现 """
        interval = self.get_parent_argv('wait_interval', interval) or 1
        return xlwait(lambda: self.find_text(dst, **kwargs), limit=limit, interval=interval)

    def waitleave_text(self, dst=None, *, limit=None, interval=None, **kwargs):
        """ 等到目标匹配文本离开（不再出现） """
        interval = self.get_parent_argv('wait_interval', interval) or 1
        return xlwait(lambda: not self.find_text(dst, **kwargs), limit=limit, interval=interval)

    def __3_查找点击功能(self):
        """ 一些常用情况的简化名称，使用起来更快捷 """

    @classmethod
    def _split_kwargs(self, kwargs):
        # 定义需要传递给 click 方法的参数
        click_params = ['x_bias', 'y_bias', 'random_bias', 'back', 'wait_change', 'wait_seconds']
        # 分离参数，click_kwargs 存储给 click 用的参数，find_img_kwargs 存储给 find_img 用的参数
        find_img_kwargs = {k: v for k, v in kwargs.items() if k not in click_params}
        click_kwargs = {k: v for k, v in kwargs.items() if k in click_params}
        return find_img_kwargs, click_kwargs

    @classmethod
    def _try_click(cls, shapes, **kwargs):
        if shapes:
            sp = shapes[0] if isinstance(shapes, list) else shapes
            sp.click(**kwargs)
            return sp

    def find_img_click(self, dst=None, **kwargs):
        find_img_kwargs, click_kwargs = self._split_kwargs(kwargs)
        shapes = self.find_img(dst, **find_img_kwargs)
        return self._try_click(shapes, **click_kwargs)

    def wait_img_click(self, dst=None, **kwargs):
        wait_img_kwargs, click_kwargs = self._split_kwargs(kwargs)
        shapes = self.wait_img(dst, **wait_img_kwargs)
        return self._try_click(shapes, **click_kwargs)

    def find_text_click(self, dst=None, **kwargs):
        find_text_kwargs, click_kwargs = self._split_kwargs(kwargs)
        shapes = self.find_text(dst, **find_text_kwargs)
        return self._try_click(shapes, **click_kwargs)

    def wait_text_click(self, dst=None, **kwargs):
        wait_text_kwargs, click_kwargs = self._split_kwargs(kwargs)
        shapes = self.wait_text(dst, **wait_text_kwargs)
        return self._try_click(shapes, **click_kwargs)

    def __4_其他高级功能(self):
        pass

    def wait_img_notchange(self, *, limit=None, interval=2, **kwargs):
        """ 一直等待到图片不再变化，以当前shot为原图，类似find_img的基础匹配 """
        interval = self.get_parent_argv('wait_interval', interval) or 1
        interval *= 2  # 这个等待可以久一点

        # 获取颜色容差参数，默认值为20
        color_tolerance = self.get_parent_argv('color_tolerance', kwargs.get('color_tolerance')) or 10
        # 获取置信度参数，默认值为0.05，表示允许5%的变化
        confidence = self.get_parent_argv('confidence', kwargs.get('confidence')) or 0.95

        # 初始截图
        prev_shot = self.shot()

        # 定义判断图片是否未变化的函数
        def check_not_change():
            nonlocal prev_shot
            current_shot = self.shot()
            # 计算图片距离
            conf = 1 - ImageTools.img_distance(prev_shot, current_shot, color_tolerance=color_tolerance)
            prev_shot = current_shot
            return conf >= confidence

        # 使用xlwait等待图片不再变化
        return xlwait(check_not_change, limit=limit, interval=interval)


class AnView(AnShape):

    def __1_构建(self):
        pass

    def __init__(self, initialdata=None, parent=None):
        """ 这个类基本都要用预设的几个特殊的init接口去初始化，实际使用中一般不直接用这个raw init
        """
        super().__init__(initialdata, parent)
        self['text'] = 'view'

        self.impath = None  # 如果是从文件读取来的数据，存储原始图片路径，在save重写回去的时候就可以缺省保存参数
        self.shapes = []  # 以列表的形式顺序存储的anshapes

    def init_from_xywh(self, xywh=None):
        if xywh is None and 'img' in self:
            xywh = [0, 0, self['img'].shape[1], self['img'].shape[0]]
        if xywh is None:
            return

        self['xywh'] = xywh
        self['center'] = [xywh[0] + xywh[2] // 2, xywh[1] + xywh[3] // 2]

        if self.get('img') is not None:
            self['pixel'] = tuple(self['img'][xywh[3] // 2, xywh[2] // 2].tolist()[::-1])

    def read_lmshapes(self, lmshapes):
        self.shapes = []
        for lmsp in lmshapes:
            sp = AnShape(parent=self)
            sp.read_lmshape(lmsp)
            self.shapes.append(sp)

    @classmethod
    def init_from_labelme_json_file(cls, lmfile):
        """ 最常用的初始化方式，从一个labelme json标注文件来初始化 """
        lmfile = XlPath(lmfile)
        lmdict = lmfile.read_json()

        view = AnView()
        view.impath = lmfile.with_name(lmdict['imagePath'])
        if not view.impath.exists():
            view.impath = lmfile.with_suffix('.jpg')

        view['img'] = xlcv.read(view.impath)
        view.read_lmshapes(lmdict['shapes'])
        view.init_from_xywh()

        return view

    @classmethod
    def init_from_image(cls, image):
        """ 从一张图片初始化
        这里默认会调用xlapi来生成labelme标注文件
        """
        # 1 读取图片
        view = AnView()
        view['img'] = xlcv.read(image)
        if isinstance(image, (str, XlPath)):
            view.impath = image

        # 2 调用ocr识别
        xlapi = get_xlapi()
        lmdict = xlapi.common_ocr(view['img'])
        view.read_lmshapes(lmdict['shapes'])
        view.init_from_xywh()

        return view

    def init_from_shot(self):
        """ 已有xywh等形状参数的情况下，可以用这个从shot初始化 """
        self['img'] = self.shot()
        xlapi = get_xlapi()
        lmdict = xlapi.common_ocr(self['img'])
        self.read_lmshapes(lmdict['shapes'])

    def __2_检索功能(self):
        pass

    def find_shapes_by_text(self, text, limit=None):
        """ 通过text内容来匹配检索shapes

        :param text: 要匹配的文本内容
        :param limit: 限制识别的数量上限
        """
        shapes = []
        for sp in self.shapes:
            if sp['text'] == text:
                shapes.append(sp)
                # 若达到数量限制则提前终止循环
                if limit is not None and len(shapes) >= limit:
                    break
        return shapes

    def find_shape_by_text(self, text):
        shapes = self.find_shapes_by_text(text, limit=1)
        return shapes[0] if shapes else None

    def loc(self, item):
        if isinstance(item, int):
            return self.shapes[item]
        elif isinstance(item, str):
            return self.find_shape_by_text(item)
        elif isinstance(item, slice):
            # 处理slice对象，返回切片后的子列表
            start, stop, step = item.indices(len(self.shapes))
            return [self.shapes[i] for i in range(start, stop, step)]
        else:
            raise TypeError

    def search(self, pattern, flags=0):
        """ 正则的search语法匹配，在shapes的sp['text']查找返回第一个的sp """
        for sp in self.shapes:
            if re.search(pattern, sp['text'], flags=flags):
                return sp

    def __getitem__(self, item):
        """
        :return:
            整数访问shapes[i]，更准确的说，是支持slice模式
            字符串则是返回第一个匹配的shape

        注意这个函数并没有__setitem__版本，如果有需要修改数据，请直接修改self.shapes
        """
        if item in self.data:
            return self.data[item]
        else:
            return self.loc(item)

    def __x_保存(self):
        pass

    def save_labelme_file(self, impath=None, save_image=True):
        """ 将数据保存为labelme标注文件

        anshape中有存储了原始的lmshape信息。

        :param impath: 要保存的路径，只要输入图片路径就行，json路径是同位置同名的，只有后缀区别
            如果已经有 self.impath ，这里可以不输入
        :param save_image: 是否覆盖原有的图片文件
            todo 注意，如果原本没有图片，这里又不进行覆盖，生成lmdict时候会报错
        :return:
        """
        # 1 读取
        impath = impath or self.impath
        if impath is None:
            return
        impath = XlPath(impath)
        jsonpath = impath.with_suffix('.json')
        os.makedirs(os.path.dirname(jsonpath), exist_ok=True)

        # 2 生成格式
        if save_image:
            xlcv.write(self['img'], impath)  # 图片有概率也发生的变动
        lmdict = LabelmeDict.gen_data(impath)
        for sp in self.shapes:
            DictTool.isub(sp, ['img'])
            shape = LabelmeDict.gen_shape(json.dumps(sp.data, ensure_ascii=False, default=str),
                                          sp['points'], sp['shape_type'],
                                          group_id=sp['group_id'], flags=sp['flags'])
            lmdict['shapes'].append(shape)

        # 3 保存
        jsonpath.write_json(lmdict, indent=2)


class AnRegion(AnShape):

    def __init__(self,
                 folder=None,
                 *,
                 read_views=False,
                 initialdata=None,
                 parent=None,
                 ):
        """
        :param folder: region标注数据所在文件夹
            有原图jpg、png和对应的json标注数据
            允许不输入，即没有目录的情况
        :param read_views: 是否直接读取目录下的views标注数据
        """
        super().__init__(initialdata, parent)
        self.folder = None if folder is None else XlPath(folder)
        self.views = {}  # views由于是文件存储，是不会重名的，所以直接用字典存储

        if read_views:  # todo 支持惰性加载？使用到对应view的时候才读取？不然以后标注文件特别多，初始化内存不爆炸？
            self.read_views()

    def _read_view(self, json_file):
        view_name = json_file.relative_to(self.folder).as_posix()[:-5]  # 注意 要去掉后缀.json
        view = AnView.init_from_labelme_json_file(json_file)
        view['text'] = view_name
        view.parent = self
        self.views[view_name] = view
        return view

    def read_views(self):
        if not (self.folder and self.folder.exists()):
            return

        self.views = {}
        for json_file in self.folder.rglob_files('*.json'):
            self._read_view(json_file)

    def loc(self, item):
        # views的具体值，采用惰性加载机制，只有使用到时，才会读取文件
        if '/' in item:
            # 注意功能效果：os.path.split('a/b/c') -> ('a/b', 'c')
            # 注意还有种很特殊的：'a' -> ('', 'a')
            view_name, shape_name = os.path.split(item)
            self._read_view(self.folder / f'{view_name}.json')
            return self.views[view_name][shape_name]
        else:
            self._read_view(self.folder / f'{item}.json')
            return self.views.get(item)

    def __getitem__(self, item):
        if item in self.data:
            return self.data[item]
        else:
            return self.loc(item)


class AnWindow(AnRegion):
    """ 某个软件的窗口区域 """

    def __0_构建(self):
        pass

    def __init__(self, folder=None, **kwargs):
        super().__init__(folder, **kwargs)

        w, h = pyautogui.size()
        self['text'] = 'screen'
        self['xywh'] = [0, 0, w, h]  # 默认获取全屏幕

        self.ctrl = None

        self._xlapi = None  # ocr工具
        self._speaker = None  # 语音播报工具
        # todo 添加数据库db？

    def set_ctrl(self, class_name=None, name=None, **kwargs):
        """ 设置窗口控件 """

        ctrl = find_ctrl(class_name=class_name, name=name, **kwargs)
        self.ctrl = UiCtrlNode(ctrl)
        self.ctrl.activate()

        self['text'] = self.ctrl.text
        self['xywh'] = self.ctrl.xywh

    @property
    def xlapi(self):
        if self._xlapi is None:
            self._xlapi = get_xlapi()
        return self._xlapi

    @property
    def speaker(self):
        if self._speaker is None:
            self._speaker = win32com.client.Dispatch('SAPI.SpVoice')
        return self.speaker

    def speak_text(self, text):
        self.speaker.Speak(text)


if __name__ == '__main__':
    pass
